@@ -3,6 +3,8 @@
 #ifndef RR_BREAKPOINT_CONDITION_H_
 #define RR_BREAKPOINT_CONDITION_H_
 
+namespace rr {
+
 class Task;
 
 class BreakpointCondition {
@@ -11,4 +13,6 @@ public:
   virtual bool evaluate(Task* t) const = 0;
 };
 
+} // namespace rr
+
 #endif // RR_BREAKPOINT_CONDITION_H_
