@@ -85,10 +85,14 @@ static void process_syscall_arch(Task* t, int syscallno) {
     case Arch::rt_sigqueueinfo:
     case Arch::rt_tgsigqueueinfo:
     case Arch::tgkill:
-    case Arch::tkill:
+    case Arch::tkill: {
       LOG(debug) << "Suppressing syscall "
                  << syscall_name(syscallno, t->arch());
+      Registers r = t->regs();
+      r.set_syscall_result(-ENOSYS);
+      t->set_regs(r);
       return;
+    }
   }
 
   LOG(debug) << "Executing syscall " << syscall_name(syscallno, t->arch());
@@ -99,6 +103,12 @@ static void process_syscall(Task* t, int syscallno){
   RR_ARCH_FUNCTION(process_syscall_arch, t->arch(), t, syscallno)
 }
 
+static void handle_ptrace_exit_event(Task *t) {
+  t->did_kill();
+  t->detach();
+  delete t;
+}
+
 /**
  * Advance execution until either a signal is received (including a SIGTRAP
  * generated by a single-step) or a syscall is made.
@@ -112,16 +122,12 @@ DiversionSession::DiversionResult DiversionSession::diversion_step(
 
   // An exit might have occurred while processing a previous syscall.
   if (t->ptrace_event() == PTRACE_EVENT_EXIT) {
+    handle_ptrace_exit_event(t);
     result.status = DIVERSION_EXITED;
     return result;
   }
 
-  // Disable syscall buffering during diversions
-  if (t->preload_globals) {
-    t->write_mem(REMOTE_PTR_FIELD(t->preload_globals, in_diversion),
-                 (unsigned char)1);
-  }
-  t->set_syscallbuf_locked(1);
+  t->set_in_diversion(true);
 
   switch (command) {
     case RUN_CONTINUE:
@@ -139,6 +145,7 @@ DiversionSession::DiversionResult DiversionSession::diversion_step(
   }
 
   if (t->ptrace_event() == PTRACE_EVENT_EXIT) {
+    handle_ptrace_exit_event(t);
     result.status = DIVERSION_EXITED;
     return result;
   }
@@ -157,6 +164,10 @@ DiversionSession::DiversionResult DiversionSession::diversion_step(
     return result;
   }
 
+  if (t->status().is_syscall()) {
+    t->apply_syscall_entry_regs();
+  }
+
   process_syscall(t, t->regs().original_syscallno());
   check_for_watchpoint_changes(t, result.break_status);
   return result;
