@@ -21,6 +21,9 @@ Event::Event(const Event& o) : event_type(o.event_type) {
     case EV_DESCHED:
       new (&Desched()) DeschedEvent(o.Desched());
       return;
+    case EV_PATCH_SYSCALL:
+      new (&PatchSyscall()) PatchSyscallEvent(o.PatchSyscall());
+      return;
     case EV_SIGNAL:
     case EV_SIGNAL_DELIVERY:
     case EV_SIGNAL_HANDLER:
@@ -43,6 +46,9 @@ Event::~Event() {
     case EV_DESCHED:
       Desched().~DeschedEvent();
       return;
+    case EV_PATCH_SYSCALL:
+      PatchSyscall().~PatchSyscallEvent();
+      return;
     case EV_SIGNAL:
     case EV_SIGNAL_DELIVERY:
     case EV_SIGNAL_HANDLER:
