@@ -148,6 +148,7 @@ enum GdbRestartType {
   RESTART_FROM_PREVIOUS,
   RESTART_FROM_EVENT,
   RESTART_FROM_CHECKPOINT,
+  RESTART_FROM_TICKS
 };
 
 enum GdbActionType { ACTION_CONTINUE, ACTION_STEP };
@@ -397,7 +398,7 @@ public:
    * target has stopped executing for some reason.  |sig| is the signal
    * that stopped execution, or 0 if execution stopped otherwise.
    */
-  void notify_stop(GdbThreadId which, int sig, uintptr_t watch_addr = 0);
+  void notify_stop(GdbThreadId which, int sig, const char *reason=nullptr);
 
   /** Notify the debugger that a restart request failed. */
   void notify_restart_failed();
@@ -579,8 +580,9 @@ public:
   const Features& features() { return features_; }
 
   enum {
-    CPU_64BIT = 0x1,
+    CPU_X86_64 = 0x1,
     CPU_AVX = 0x2,
+    CPU_AARCH64 = 0x4
   };
   void set_cpu_features(uint32_t features) { cpu_features_ = features; }
   uint32_t cpu_features() const { return cpu_features_; }
@@ -598,6 +600,9 @@ public:
   */
   bool is_connection_alive();
 
+  bool hwbreak_supported() { return hwbreak_supported_; }
+  bool swbreak_supported() { return swbreak_supported_; }
+
 private:
   /**
    * read() incoming data exactly one time, successfully.  May block.
@@ -665,7 +670,7 @@ private:
   bool process_packet();
   void consume_request();
   void send_stop_reply_packet(GdbThreadId thread, int sig,
-                              uintptr_t watch_addr = 0);
+                              const char *reason);
   void send_file_error_reply(int system_errno);
 
   // Current request to be processed.
@@ -689,6 +694,8 @@ private:
   Features features_;
   bool connection_alive_;
   bool multiprocess_supported_; // client supports multiprocess extension
+  bool hwbreak_supported_; // client supports hwbreak extension
+  bool swbreak_supported_; // client supports swbreak extension
 };
 
 } // namespace rr
