@@ -87,13 +87,19 @@ void MmappedFileMonitor::did_write(Task* t, const std::vector<Range>& ranges,
             m.emu_file->ensure_size(local_offset + r.length);
           } else {
             ASSERT(t, !v->task_set().empty());
-            Task* tt = *v->task_set().begin();
             // We will record multiple writes if the file is mapped multiple
             // times. This is inefficient --- one is sufficient --- but not
             // wrong.
             // Make sure we use a task for this address space. `t` might have
             // a different address space.
-            static_cast<RecordTask*>(tt)->record_remote(km.intersect(mr));
+            for (auto tt : v->task_set()) {
+              // If the task here has execed, we may not be able to record its
+              // memory any longer, so loop through all tasks in this address
+              // space in turn in case any *didn't* exec.
+              if (static_cast<RecordTask*>(tt)->record_remote_fallible(km.intersect(mr)) > 0) {
+                break;
+              }
+            }
           }
         }
         local_offset += r.length;
