@@ -6,7 +6,8 @@
 
 #include "AutoRemoteSyscalls.h"
 #include "RecordSession.h"
-#include "RecordTask.h"
+#include "ReplaySession.h"
+#include "ReplayTask.h"
 #include "log.h"
 
 using namespace std;
@@ -32,16 +33,33 @@ ProcMemMonitor::ProcMemMonitor(Task* t, const string& pathname) {
 
 void ProcMemMonitor::did_write(Task* t, const std::vector<Range>& ranges,
                                LazyOffset& lazy_offset) {
-  if (t->session().is_replaying() || ranges.empty()) {
+  if (ranges.empty()) {
     return;
   }
-  auto* target = static_cast<RecordTask*>(t->session().find_task(tuid));
+  int64_t offset = lazy_offset.retrieve(true);
+
+  // In prior versions of rr, we recorded this directly into the trace.
+  // If so, there's nothing to do here.
+  if (t->session().is_replaying() && t->session().as_replay()->explicit_proc_mem()) {
+    return;
+  }
+
+  if (t->session().is_recording()) {
+    // Nothing to do now (though we may have just recorded the offset)
+    return;
+  }
+
+  auto* target = static_cast<ReplayTask*>(t->session().find_task(tuid));
   if (!target) {
     return;
   }
-  int64_t offset = lazy_offset.retrieve(false);
+
   for (auto& r : ranges) {
-    target->record_remote(remote_ptr<void>(offset), r.length);
+    auto bytes = t->read_mem(r.data.cast<uint8_t>(), r.length);
+    remote_ptr<uint8_t> target_addr = offset;
+    target->write_mem(target_addr, bytes.data(), r.length);
+    target->vm()->maybe_update_breakpoints(target, target_addr,
+                                           r.length);
     offset += r.length;
   }
 }
