@@ -25,9 +25,9 @@ void ReplayTimeline::InternalMark::full_print(FILE* out) const {
   fprintf(out, "{%s,regs:", buf.str().c_str());
   proto.regs.print_register_file(out);
   fputs(",return_addresses=[", out);
-  for (int i = 0; i < ReturnAddressList::COUNT; ++i) {
+  for (int i = 0; i < ReturnAddressList::MAX_COUNT; ++i) {
     fprintf(out, "%p", (void*)proto.return_addresses.addresses[i].as_int());
-    if (i + 1 < ReturnAddressList::COUNT) {
+    if (i + 1 < ReturnAddressList::MAX_COUNT) {
       fputc(',', out);
     }
   }
@@ -80,7 +80,7 @@ bool ReplayTimeline::less_than(const Mark& m1, const Mark& m2) {
 ReplayTimeline::ReplayTimeline(std::shared_ptr<ReplaySession> session)
     : current(std::move(session)),
       breakpoints_applied(false),
-      reverse_execution_barrier_event(0) {
+      reverse_execution_barrier_event_(0) {
   current->set_visible_execution(false);
 }
 
@@ -420,6 +420,38 @@ void ReplayTimeline::update_strategy_and_fix_watchpoint_quirk(
   }
 }
 
+void ReplayTimeline::seek_to_ticks(FrameTime time, Ticks ticks) {
+  seek_to_before_key(MarkKey(time, ticks, ReplayStepKey()));
+  unapply_breakpoints_and_watchpoints();
+
+  while (current->trace_reader().time() < time) {
+    ReplaySession::StepConstraints constraints(RUN_CONTINUE);
+    constraints.stop_at_time = time;
+    ReplayResult result = current->replay_step(constraints);
+    if (result.status != REPLAY_CONTINUE) {
+      FATAL() << "Trace finished before target was reached";
+    }
+  }
+
+  ReplayTask* t = current->current_task();
+  if (t->tick_count() < ticks) {
+    // First step using the ticks counter until we enter the skid region
+    ReplaySession::StepConstraints constraints(RUN_CONTINUE);
+    while (true) {
+      constraints.ticks_target = ticks;
+      ReplayResult result = current->replay_step(constraints);
+      if (result.break_status.approaching_ticks_target) {
+        break;
+      }
+    }
+
+    // Then singlestep until we hit our ticks target
+    while (t->tick_count() < ticks) {
+      current->replay_step(RUN_SINGLESTEP_FAST_FORWARD);
+    }
+  }
+}
+
 ReplayResult ReplayTimeline::replay_step_to_mark(
     const Mark& mark, ReplayStepToMarkStrategy& strategy) {
   ReplayTask* t = current->current_task();
@@ -778,7 +810,7 @@ ReplayResult ReplayTimeline::singlestep_with_breakpoints_disabled() {
 }
 
 bool ReplayTimeline::is_start_of_reverse_execution_barrier_event() {
-  if (current->trace_reader().time() != reverse_execution_barrier_event ||
+  if (current->trace_reader().time() != reverse_execution_barrier_event_ ||
       current->current_step_key().in_execution()) {
     return false;
   }
@@ -900,6 +932,19 @@ static ReplayTask* to_replay_task(const BreakStatus& status) {
   return static_cast<ReplayTask*>(status.task);
 }
 
+static bool arch_watch_fires_before_instr(SupportedArch arch) {
+  switch (arch) {
+    case x86:
+    case x86_64:
+      return false;
+    case aarch64:
+      return true;
+    default:
+      FATAL() << "Unknown architecture";
+      return false;
+  }
+}
+
 ReplayResult ReplayTimeline::reverse_continue(
     const std::function<bool(ReplayTask* t)>& stop_filter,
     const std::function<bool()>& interrupt_check) {
@@ -946,17 +991,23 @@ ReplayResult ReplayTimeline::reverse_continue(
                << " up to " << end;
 
     bool at_breakpoint = false;
+    bool before_watchpoint = false;
     ReplayStepToMarkStrategy strategy;
     int stop_count = 0;
     bool made_progress_between_stops = false;
     remote_code_ptr avoidable_stop_ip;
     Ticks avoidable_stop_ticks = 0;
     while (true) {
-      apply_breakpoints_and_watchpoints();
       ReplayResult result;
       if (at_breakpoint) {
         result = singlestep_with_breakpoints_disabled();
+      } else if (before_watchpoint) {
+        // N.B.: This state is only reached on architectures where watchpoints
+        // fire *before* applying instruction effects (e.g. aarch64)
+        unapply_breakpoints_and_watchpoints();
+        result = current->replay_step(RUN_SINGLESTEP);
       } else {
+        apply_breakpoints_and_watchpoints();
         result = replay_step_to_mark(end, strategy);
         // This will remove all reverse-exec checkpoints ahead of the
         // current time, and add new ones if necessary. This should be
@@ -997,6 +1048,9 @@ ReplayResult ReplayTimeline::reverse_continue(
         } else {
           LOG(debug) << "Found watch break at " << dest << ", addr="
                      << result.break_status.data_watchpoints_hit()[0].addr;
+          if (arch_watch_fires_before_instr(current->arch())) {
+            before_watchpoint = true;
+          }
         }
         final_result = result;
         final_tuid = result.break_status.task ? result.break_status.task->tuid()
@@ -1112,7 +1166,7 @@ ReplayResult ReplayTimeline::reverse_singlestep(
   while (true) {
     Mark end = outer;
     Mark start;
-    bool seen_barrier;
+    bool seen_barrier = false;
 
     while (true) {
       MarkKey current_key = end.ptr->proto.key;
@@ -1242,6 +1296,14 @@ ReplayResult ReplayTimeline::reverse_singlestep(
               // This last step is not usable.
               LOG(debug) << "   not usable, stopping now";
               break;
+            } else if (now == end &&
+                       result.break_status.signal &&
+                       result.break_status.signal->si_signo == SIGTRAP &&
+                       is_advanced_pc_and_signaled_instruction(result.break_status.task,
+                                                               result.break_status.task->ip())) {
+              LOG(debug) << "   singlestepped exactly to instruction that advances pc and signals (e.g. int3),"
+                         << " pretending we stopped earlier.";
+              break;
             }
             destination_candidate = step_start;
             LOG(debug) << "Setting candidate after step: "
