@@ -43,6 +43,7 @@ RerunCommand RerunCommand::singleton(
     "  -f, --function=<ADDR>      when starting tracing, push sentinel return\n"
     "                             address and jump to <ADDR> to fake call\n"
     "  --singlestep=<REGS>        dump <REGS> after each singlestep\n"
+    "  --event-regs=<REGS>        dump <REGS> after each event\n"
     "  -r, --raw                  dump registers in raw format\n"
     "  -s, --trace-start=<EVENT>  start tracing at <EVENT>\n"
     "  -u, --cpu-unbound          allow replay to run on any CPU. Default is\n"
@@ -73,6 +74,8 @@ enum TraceFieldKind {
   TRACE_GP_REG,            // outputs 64-bit value
   TRACE_XMM_REG,           // outputs 128-bit value
   TRACE_YMM_REG,           // outputs 256-bit value
+  TRACE_FIP,               // outputs 64-bit value
+  TRACE_TID,               // outputs 32-bit value
 };
 struct TraceField {
   TraceFieldKind kind;
@@ -84,6 +87,7 @@ struct RerunFlags {
   FrameTime trace_end;
   remote_code_ptr function;
   vector<TraceField> singlestep_trace;
+  vector<TraceField> event_trace;
   bool raw;
   bool cpu_unbound;
 
@@ -112,6 +116,30 @@ static uint8_t user_regs_fields[16] = {
   offsetof(user_regs_struct, esp), offsetof(user_regs_struct, ebp),
   offsetof(user_regs_struct, esi), offsetof(user_regs_struct, edi),
 };
+#elif defined(__aarch64__)
+#define user_regs_struct NativeArch::user_regs_struct
+static uint16_t user_regs_fields[34] = {
+  offsetof(user_regs_struct, x[0]), offsetof(user_regs_struct, x[1]),
+  offsetof(user_regs_struct, x[2]), offsetof(user_regs_struct, x[3]),
+  offsetof(user_regs_struct, x[4]), offsetof(user_regs_struct, x[5]),
+  offsetof(user_regs_struct, x[6]), offsetof(user_regs_struct, x[7]),
+  offsetof(user_regs_struct, x[8]), offsetof(user_regs_struct, x[9]),
+  offsetof(user_regs_struct, x[10]), offsetof(user_regs_struct, x[11]),
+  offsetof(user_regs_struct, x[12]), offsetof(user_regs_struct, x[13]),
+  offsetof(user_regs_struct, x[14]), offsetof(user_regs_struct, x[15]),
+  offsetof(user_regs_struct, x[16]), offsetof(user_regs_struct, x[17]),
+  offsetof(user_regs_struct, x[18]), offsetof(user_regs_struct, x[19]),
+  offsetof(user_regs_struct, x[20]), offsetof(user_regs_struct, x[21]),
+  offsetof(user_regs_struct, x[22]), offsetof(user_regs_struct, x[23]),
+  offsetof(user_regs_struct, x[24]), offsetof(user_regs_struct, x[25]),
+  offsetof(user_regs_struct, x[26]), offsetof(user_regs_struct, x[27]),
+  offsetof(user_regs_struct, x[28]), offsetof(user_regs_struct, x[29]),
+  offsetof(user_regs_struct, x[30]),
+  offsetof(user_regs_struct, sp),
+  offsetof(user_regs_struct, pc),
+  offsetof(user_regs_struct, pstate)
+};
+#undef user_regs_struct
 #else
 #error Unsupported architecture
 #endif
@@ -166,13 +194,15 @@ static uint64_t seg_reg(const Registers& regs, uint8_t index) {
 }
 
 static void print_regs(Task* t, FrameTime event, uint64_t instruction_count,
-                       const RerunFlags& flags, FILE* out) {
+                       const RerunFlags& flags, const vector<TraceField>& fields, FILE* out) {
+  if (fields.empty()) {
+    return;
+  }
   union {
-    struct user_regs_struct gp_regs;
+    NativeArch::user_regs_struct gp_regs;
     uintptr_t regs_values[sizeof(struct user_regs_struct) / sizeof(uintptr_t)];
   };
   bool got_gp_regs = false;
-  const vector<TraceField>& fields = flags.singlestep_trace;
   bool first = true;
 
   for (auto& field : fields) {
@@ -269,6 +299,8 @@ static void print_regs(Task* t, FrameTime event, uint64_t instruction_count,
               memset(value, 0, sizeof(value));
             }
             break;
+          default:
+            FATAL() << "Unexpected architecture";
         }
         char buf[8];
         sprintf(buf, "xmm%d", field.reg_num);
@@ -301,12 +333,23 @@ static void print_regs(Task* t, FrameTime event, uint64_t instruction_count,
               memset(value, 0, sizeof(value));
             }
             break;
+          default:
+            FATAL() << "Unexpected architecture";
         }
         char buf[8];
         sprintf(buf, "ymm%d", field.reg_num);
         print_value(buf, value, sizeof(value), flags, out);
         break;
       }
+      case TRACE_FIP: {
+        bool defined;
+        uint64_t value = t->extra_regs().read_fip(&defined);
+        print_value("fip", &value, sizeof(value), flags, out);
+        break;
+      }
+      case TRACE_TID:
+        print_value("tid", &t->rec_tid, sizeof(t->rec_tid), flags, out);
+        break;
     }
   }
 
@@ -379,6 +422,10 @@ static bool parse_regs(const string& value, vector<TraceField>* out) {
       out->push_back({ TRACE_SEG_REG, (uint8_t)find_seg_reg(reg) });
     } else if (reg == "xinuse") {
       out->push_back({ TRACE_XINUSE, 0 });
+    } else if (reg == "fip") {
+      out->push_back({ TRACE_FIP, 0 });
+    } else if (reg == "tid") {
+      out->push_back({ TRACE_TID, 0 });
     } else {
       fprintf(stderr, "Unknown register '%s'\n", reg.c_str());
       return false;
@@ -394,6 +441,7 @@ static bool parse_rerun_arg(vector<string>& args, RerunFlags& flags) {
 
   static const OptionSpec options[] = {
     { 1, "singlestep", HAS_PARAMETER },
+    { 2, "event-regs", HAS_PARAMETER },
     { 'e', "trace-end", HAS_PARAMETER },
     { 'f', "function", HAS_PARAMETER },
     { 'r', "raw", NO_PARAMETER },
@@ -411,6 +459,11 @@ static bool parse_rerun_arg(vector<string>& args, RerunFlags& flags) {
         return false;
       }
       break;
+    case 2:
+      if (!parse_regs(opt.value, &flags.event_trace)) {
+        return false;
+      }
+      break;
     case 'e':
       if (!opt.verify_valid_int(1, UINT32_MAX)) {
         return false;
@@ -471,28 +524,6 @@ static bool ignore_singlestep_for_event(const Event& ev) {
   }
 }
 
-/**
- * In KVM virtual machines (and maybe others), singlestepping over CPUID
- * executes the following instruction as well. Work around that.
- */
-static bool maybe_set_breakpoint_after_cpuid(Task* t) {
-  if (!t) {
-    return false;
-  }
-  uint8_t bytes[2];
-  if (t->read_bytes_fallible(t->ip().to_data_ptr<void>(), 2, bytes) != 2) {
-    return false;
-  }
-  if (bytes[0] != 0x0f || bytes[1] != 0xa2) {
-    return false;
-  }
-  return t->vm()->add_breakpoint(t->ip() + 2, BKPT_USER);
-}
-
-static void clear_breakpoint_after_cpuid(Task* t) {
-  t->vm()->remove_breakpoint(t->ip(), BKPT_USER);
-}
-
 static const uint64_t sentinel_ret_address = 9;
 
 static void run_diversion_function(ReplaySession& replay, Task* task,
@@ -514,7 +545,7 @@ static void run_diversion_function(ReplaySession& replay, Task* task,
   while (true) {
     DiversionSession::DiversionResult result =
         diversion_session->diversion_step(t, cmd);
-    print_regs(t, 0, 0, flags, stdout);
+    print_regs(t, 0, 0, flags, flags.singlestep_trace, stdout);
     if (result.break_status.signal) {
       if (result.break_status.signal->si_signo == SIGSEGV &&
           result.break_status.signal->si_addr == (void*)sentinel_ret_address) {
@@ -538,6 +569,12 @@ static int rerun(const string& trace_dir, const RerunFlags& flags) {
   ReplaySession::shr_ptr replay_session = ReplaySession::create(trace_dir, session_flags(flags));
   uint64_t instruction_count_within_event = 0;
   bool done_first_step = false;
+  bool need_to_singlestep = !flags.singlestep_trace.empty();
+  for (auto& v : flags.event_trace) {
+    if (v.kind == TRACE_INSTRUCTION_COUNT) {
+      need_to_singlestep = true;
+    }
+  }
 
   // Now that we've spawned the replay, raise our resource limits if
   // possible.
@@ -547,6 +584,7 @@ static int rerun(const string& trace_dir, const RerunFlags& flags) {
     RunCommand cmd = RUN_CONTINUE;
 
     Task* old_task = replay_session->current_task();
+    auto old_task_tuid = old_task ? old_task->tuid() : TaskUid();
     remote_code_ptr old_ip = old_task ? old_task->ip() : remote_code_ptr();
     FrameTime before_time = replay_session->trace_reader().time();
     if (replay_session->done_initial_exec() &&
@@ -557,36 +595,28 @@ static int rerun(const string& trace_dir, const RerunFlags& flags) {
           return 0;
         }
 
-        if (!flags.singlestep_trace.empty()) {
-          done_first_step = true;
-          print_regs(old_task, before_time - 1, instruction_count_within_event,
-                     flags, stdout);
-        }
+        done_first_step = true;
+        print_regs(old_task, before_time - 1, instruction_count_within_event,
+                   flags, flags.singlestep_trace, stdout);
       }
 
-      cmd = RUN_SINGLESTEP_FAST_FORWARD;
+      if (need_to_singlestep) {
+        cmd = RUN_SINGLESTEP_FAST_FORWARD;
+      }
     }
 
     Event replayed_event = replay_session->current_trace_frame().event();
 
-    bool set_breakpoint = maybe_set_breakpoint_after_cpuid(old_task);
     auto result = replay_session->replay_step(cmd);
-    if (set_breakpoint) {
-      clear_breakpoint_after_cpuid(old_task);
-      if (result.break_status.breakpoint_hit ||
-          result.break_status.singlestep_complete) {
-        ASSERT(old_task, old_task->ip() == old_ip + 2);
-        result.break_status.breakpoint_hit = false;
-        result.break_status.singlestep_complete = true;
-      }
-    }
-
     if (result.status == REPLAY_EXITED) {
       break;
     }
 
     FrameTime after_time = replay_session->trace_reader().time();
     if (cmd != RUN_CONTINUE) {
+      // The old_task may have exited (and been deallocated) in the `replay_session->replay_step(cmd)` above.
+      // So we need to try and obtain it from the session again to make sure it still exists.
+      Task* old_task = old_task_tuid.tid() ? replay_session->find_task(old_task_tuid) : nullptr;
       remote_code_ptr after_ip = old_task ? old_task->ip() : remote_code_ptr();
       DEBUG_ASSERT(after_time >= before_time && after_time <= before_time + 1);
 
@@ -604,15 +634,14 @@ static int rerun(const string& trace_dir, const RerunFlags& flags) {
           // event
           (!ignore_singlestep_for_event(replayed_event) ||
            before_time == after_time) &&
-          (!result.did_fast_forward || old_ip != after_ip ||
+          (!result.incomplete_fast_forward || old_ip != after_ip ||
            before_time < after_time);
-      if (!flags.singlestep_trace.empty() &&
-          cmd == RUN_SINGLESTEP_FAST_FORWARD &&
+      if (cmd == RUN_SINGLESTEP_FAST_FORWARD &&
           (singlestep_really_complete ||
            (before_time < after_time &&
             treat_event_completion_as_singlestep_complete(replayed_event)))) {
         print_regs(old_task, before_time, instruction_count_within_event, flags,
-                   stdout);
+                   flags.singlestep_trace, stdout);
       }
 
       if (singlestep_really_complete) {
@@ -622,6 +651,8 @@ static int rerun(const string& trace_dir, const RerunFlags& flags) {
     if (before_time < after_time) {
       LOG(debug) << "Completed event " << before_time
                  << " instruction_count=" << instruction_count_within_event;
+      print_regs(old_task, before_time, instruction_count_within_event, flags,
+                 flags.event_trace, stdout);
       instruction_count_within_event = 1;
     }
   }
