@@ -8,6 +8,7 @@
 #include <unordered_map>
 #include <unordered_set>
 #include <vector>
+#include <map>
 
 #include "Command.h"
 #include "ElfReader.h"
@@ -31,6 +32,11 @@ namespace rr {
 ///   external file, one of "debuglink", "debugaltlink". Note that for "debugaltlink", it is possible
 ///   to have the same file appearing multiple times with different build-ids, when it's shared by
 ///   multiple ELF binaries.
+/// "dwo": an array of objects, {"name":<name>, "trace_file":<name>, "comp_dir":<path>, "id":<value>}
+///   These are the references to DWO files found in the trace binaries. "name" is the value of
+/// DW_AT_GNU_dwo_name. "trace_file" is the trace-relative binary file name. "comp_dir" is the
+/// value of DW_AT_comp_dir for the compilation unit containing the DWO reference. "id"
+/// is the value of DW_AT_GNU_dwo_id (64 bit number).
 /// "symlinks": an array of objects, {"from":<path>, "to":<path>}.
 ///   These symlinks that exist in the filesystem that are relevant to the source file paths.
 /// "files": a map from VCS directory name to array of source files relative to that directory
@@ -45,7 +51,33 @@ protected:
   static SourcesCommand singleton;
 };
 
-SourcesCommand SourcesCommand::singleton("sources", " rr sources [<trace_dir>]\n");
+SourcesCommand SourcesCommand::singleton(
+    "sources",
+    " rr sources [<trace_dir>]\n"
+    "  --substitute=LIBRARY=PATH  When searching for the source to LIBRARY,\n"
+    "                             substitute PATH in place of the path stored\n"
+    "                             in the library's DW_AT_comp_dir property\n"
+    "                             for all compilation units.\n");
+
+class ExplicitSourcesCommand : public Command {
+public:
+  virtual int run(vector<string>& args) override;
+
+protected:
+  ExplicitSourcesCommand(const char* name, const char* help) : Command(name, help) {}
+
+  static ExplicitSourcesCommand singleton;
+};
+
+ExplicitSourcesCommand ExplicitSourcesCommand::singleton(
+    "explicit-sources",
+    " rr explicit-sources [<file>...]\n"
+    "  Like `rr sources` but instead of scanning the binary files used in a\n"
+    "  trace, scans an explicit list of files.\n"
+    "  --substitute=LIBRARY=PATH  When searching for the source to LIBRARY,\n"
+    "                             substitute PATH in place of the path stored\n"
+    "                             in the library's DW_AT_comp_dir property\n"
+    "                             for all compilation units.\n");
 
 static void parent_dir(string& s) {
   size_t p = s.rfind('/');
@@ -63,47 +95,6 @@ static void base_name(string& s) {
   }
 }
 
-static bool is_component(const char* p, const char* component) {
-  while (*component) {
-    if (*p != *component) {
-      return false;
-    }
-    ++p;
-    ++component;
-  }
-  return *p == '/' || !*p;
-}
-
-static void normalize_file_name(string& s) {
-  size_t s_len = s.size();
-  size_t out = 0;
-  for (size_t i = 0; i < s_len; ++i) {
-    if (s[i] == '/') {
-      if (s.c_str()[i + 1] == '/') {
-        // Skip redundant '/'
-        continue;
-      }
-      if (is_component(s.c_str() + i + 1, ".")) {
-        // Skip redundant '/.'
-        ++i;
-        continue;
-      }
-      if (is_component(s.c_str() + i + 1, "..")) {
-        // Peel off '/..'
-        size_t p = s.rfind('/', out - 1);
-        if (p != string::npos) {
-          out = p;
-          i += 2;
-          continue;
-        }
-      }
-    }
-    s[out] = s[i];
-    ++out;
-  }
-  s.resize(out);
-}
-
 // file_name cannot be null, but the others can be.
 static void resolve_file_name(const char* original_file_dir,
                               const char* comp_dir, const char* rel_dir,
@@ -128,17 +119,36 @@ static void resolve_file_name(const char* original_file_dir,
   file_names->insert(move(s));
 }
 
-static bool list_source_files(ElfFileReader& reader, const string& original_file_name,
-                              set<string>* file_names) {
+struct DwoInfo {
+  string name;
+  string trace_file;
+  // Could be an empty string
+  string comp_dir;
+  uint64_t id;
+};
+
+static bool process_compilation_units(ElfFileReader& reader,
+                                      const string& trace_relative_name,
+                                      const string& original_file_name,
+                                      const string& comp_dir_substitution,
+                                      set<string>* file_names, vector<DwoInfo>* dwos) {
   DwarfSpan debug_info = reader.dwarf_section(".debug_info");
   DwarfSpan debug_abbrev = reader.dwarf_section(".debug_abbrev");
   DwarfSpan debug_str = reader.dwarf_section(".debug_str");
+  DwarfSpan debug_str_offsets = reader.dwarf_section(".debug_str_offsets");
   DwarfSpan debug_line = reader.dwarf_section(".debug_line");
+  DwarfSpan debug_line_str = reader.dwarf_section(".debug_line_str");
   if (debug_info.empty() || debug_abbrev.empty() ||
       debug_str.empty() || debug_line.empty())  {
     return false;
   }
 
+  DebugStrSpans debug_strs = {
+    debug_str,
+    debug_str_offsets,
+    debug_line_str,
+  };
+
   string original_file_dir = original_file_name;
   parent_dir(original_file_dir);
 
@@ -149,22 +159,65 @@ static bool list_source_files(ElfFileReader& reader, const string& original_file
     if (!ok) {
       break;
     }
-    const char* comp_dir = cu.die().string_attr(DW_AT_comp_dir, debug_str, &ok);
+    int64_t str_offsets_base = cu.die().section_ptr_attr(DW_AT_str_offsets_base, &ok);
     if (!ok) {
       continue;
     }
-    const char* source_file_name = cu.die().string_attr(DW_AT_name, debug_str, &ok);
+    if (str_offsets_base > 0) {
+      cu.set_str_offsets_base(str_offsets_base);
+    } else {
+      cu.set_str_offsets_base(0);
+    }
+    const char* comp_dir;
+    if (!comp_dir_substitution.empty()) {
+      comp_dir = comp_dir_substitution.c_str();
+    } else {
+      comp_dir = cu.die().string_attr(cu, DW_AT_comp_dir, debug_strs, &ok);
+      if (!ok) {
+        continue;
+      }
+    }
+    const char* dwo_name = cu.die().string_attr(cu, DW_AT_GNU_dwo_name, debug_strs, &ok);
+    if (!ok || !dwo_name) {
+      dwo_name = cu.die().string_attr(cu, DW_AT_dwo_name, debug_strs, &ok);
+      if (!ok) {
+        continue;
+      }
+    }
+    if (dwo_name) {
+      bool has_dwo_id = false;
+      uint64_t dwo_id = cu.dwo_id();
+      if (dwo_id != 0) {
+        has_dwo_id = true;
+      }
+      if (!has_dwo_id) {
+        dwo_id = cu.die().unsigned_attr(DW_AT_GNU_dwo_id, &has_dwo_id, &ok);
+        if (!ok) {
+          continue;
+        }
+      }
+      if (has_dwo_id) {
+        string c;
+        if (comp_dir) {
+          c = comp_dir;
+        }
+        dwos->push_back({ dwo_name, trace_relative_name, move(c), dwo_id });
+      } else {
+        LOG(warn) << "DW_AT_GNU_dwo_name but not DW_AT_GNU_dwo_id";
+      }
+    }
+    const char* source_file_name = cu.die().string_attr(cu, DW_AT_name, debug_strs, &ok);
     if (!ok) {
       continue;
     }
     if (source_file_name) {
       resolve_file_name(original_file_dir.c_str(), comp_dir, nullptr, source_file_name, file_names);
     }
-    intptr_t stmt_list = cu.die().lineptr_attr(DW_AT_stmt_list, &ok);
+    intptr_t stmt_list = cu.die().section_ptr_attr(DW_AT_stmt_list, &ok);
     if (stmt_list < 0 || !ok) {
       continue;
     }
-    DwarfLineNumberTable lines(debug_line.subspan(stmt_list), &ok);
+    DwarfLineNumberTable lines(cu, debug_line.subspan(stmt_list), debug_strs, &ok);
     if (!ok) {
       continue;
     }
@@ -202,9 +255,12 @@ struct ExternalDebugInfo {
   }
 };
 
-static bool try_auxiliary_file(ElfFileReader& trace_file_reader, const string& original_file_name,
+static bool try_auxiliary_file(ElfFileReader& trace_file_reader,
+                               const string& trace_relative_name,
+                               const string& original_file_name,
                                set<string>* file_names, const string& aux_file_name,
                                const char* file_type,
+                               vector<DwoInfo>* dwos,
                                set<ExternalDebugInfo>* external_debug_info) {
   if (aux_file_name.empty()) {
     return false;
@@ -235,7 +291,7 @@ static bool try_auxiliary_file(ElfFileReader& trace_file_reader, const string& o
     LOG(warn) << "Main ELF binary has no build ID!";
     return false;
   }
-  if (!list_source_files(reader, original_file_name, file_names)) {
+  if (!process_compilation_units(reader, trace_relative_name, original_file_name, {}, file_names, dwos)) {
     LOG(warn) << "No debuginfo!";
     return false;
   }
@@ -412,56 +468,47 @@ static bool starts_with(const string& s, const string& prefix) {
   return strncmp(s.c_str(), prefix.c_str(), prefix.size()) == 0;
 }
 
-static int sources(const string& trace_dir) {
-  TraceReader trace(trace_dir);
-  DIR* files = opendir(trace.dir().c_str());
-  if (!files) {
-    FATAL() << "Can't open trace dir";
-  }
-
-  // (Trace file name, original file name) pairs
-  map<string, string> binary_file_names;
-  while (true) {
-    TraceReader::MappedData data;
-    bool found;
-    KernelMapping km = trace.read_mapped_region(
-        &data, &found, TraceReader::VALIDATE, TraceReader::ANY_TIME);
-    if (!found) {
-      break;
-    }
-    if (data.source == TraceReader::SOURCE_FILE) {
-      binary_file_names.insert(make_pair(move(data.file_name), km.fsname()));
-    }
-  }
-
+static int sources(const map<string, string>& binary_file_names, const map<string, string>& comp_dir_substitutions, bool is_explicit) {
   vector<string> relevant_binary_names;
   set<string> file_names;
   set<ExternalDebugInfo> external_debug_info;
+  vector<DwoInfo> dwos;
   for (auto& pair : binary_file_names) {
-    ScopedFd fd(pair.first.c_str(), O_RDONLY);
+    string trace_relative_name = pair.first;
+    string original_name = pair.second;
+    const char* file_name = is_explicit ? original_name.c_str() : trace_relative_name.c_str();
+    ScopedFd fd(file_name, O_RDONLY);
     if (!fd.is_open()) {
-      FATAL() << "Can't open " << pair.first;
+      FATAL() << "Can't open " << file_name;
     }
-    LOG(info) << "Examining " << pair.first;
+    LOG(info) << "Examining " << file_name;
     ElfFileReader reader(fd);
     if (!reader.ok()) {
       LOG(info) << "Probably not an ELF file, skipping";
       continue;
     }
-    bool has_source_files = list_source_files(reader, pair.second, &file_names);
+    if (!is_explicit) {
+      base_name(trace_relative_name);
+    }
+    base_name(original_name);
+    bool has_source_files;
+    auto it = comp_dir_substitutions.find(original_name);
+    if (it != comp_dir_substitutions.end()) {
+      has_source_files = process_compilation_units(reader, trace_relative_name, pair.second, it->second, &file_names, &dwos);
+    } else {
+      has_source_files = process_compilation_units(reader, trace_relative_name, pair.second, {}, &file_names, &dwos);
+    }
 
     Debuglink debuglink = reader.read_debuglink();
-    has_source_files |= try_auxiliary_file(reader, pair.second,
-      &file_names, debuglink.file_name, "debuglink", &external_debug_info);
+    has_source_files |= try_auxiliary_file(reader, trace_relative_name, pair.second,
+      &file_names, debuglink.file_name, "debuglink", &dwos, &external_debug_info);
 
     Debugaltlink debugaltlink = reader.read_debugaltlink();
-    has_source_files |= try_auxiliary_file(reader, pair.second,
-      &file_names, debugaltlink.file_name, "debugaltlink", &external_debug_info);
+    has_source_files |= try_auxiliary_file(reader, trace_relative_name, pair.second,
+      &file_names, debugaltlink.file_name, "debugaltlink", &dwos, &external_debug_info);
 
     if (has_source_files) {
-      string name = pair.first;
-      base_name(name);
-      relevant_binary_names.push_back(move(name));
+      relevant_binary_names.push_back(move(trace_relative_name));
     } else {
       LOG(info) << "No debuginfo found";
     }
@@ -517,6 +564,21 @@ static int sources(const string& trace_dir) {
     ++index;
   }
   printf("  ],\n");
+  printf("  \"dwos\":[\n");
+  index = 0;
+  for (auto& d : dwos) {
+    printf("    { \"name\":\"%s\", \"trace_file\":\"%s\", ",
+           json_escape(d.name).c_str(),
+           json_escape(d.trace_file).c_str());
+    if (!d.comp_dir.empty()) {
+      printf("\"comp_dir\":\"%s\", ", json_escape(d.comp_dir).c_str());
+    }
+    printf("\"id\":%llu }%s\n",
+           (unsigned long long)d.id,
+           index == dwos.size() - 1 ? "" : ",");
+    ++index;
+  }
+  printf("  ],\n");
   printf("  \"symlinks\":[\n");
   for (size_t i = 0; i < symlinks.size(); ++i) {
     auto& link = symlinks[i];
@@ -548,17 +610,103 @@ static int sources(const string& trace_dir) {
   return 0;
 }
 
-int SourcesCommand::run(vector<string>& args) {
-  while (parse_global_option(args)) {
+static bool parse_sources_option(vector<string>& args, map<string, string>& comp_dir_substitutions) {
+  if (parse_global_option(args)) {
+    return true;
   }
 
+  static const OptionSpec options[] = {
+    { 0, "substitute", HAS_PARAMETER }
+  };
+
+  ParsedOption opt;
+  if (!Command::parse_option(args, options, &opt)) {
+    return false;
+  }
+
+  switch (opt.short_name) {
+    case 0: {
+      auto pos = opt.value.find_first_of('=');
+      if (pos != string::npos) {
+        auto k = opt.value.substr(0, pos);
+        auto v = opt.value.substr(pos+1);
+        comp_dir_substitutions.insert(std::pair<string, string>(k, v));
+      }
+      break;
+    }
+  }
+
+  return true;
+}
+
+int SourcesCommand::run(vector<string>& args) {
+  map<string, string> comp_dir_substitutions;
+  while (parse_sources_option(args, comp_dir_substitutions)) {
+  }
+
+  // (Trace file name, original file name) pairs
   string trace_dir;
   if (!parse_optional_trace_dir(args, &trace_dir)) {
     print_help(stderr);
     return 1;
   }
 
-  return sources(trace_dir);
+  TraceReader trace(trace_dir);
+  DIR* files = opendir(trace.dir().c_str());
+  if (!files) {
+    FATAL() << "Can't open trace dir";
+  }
+
+  map<string, string> binary_file_names;
+  while (true) {
+    TraceReader::MappedData data;
+    bool found;
+    KernelMapping km = trace.read_mapped_region(
+        &data, &found, TraceReader::VALIDATE, TraceReader::ANY_TIME);
+    if (!found) {
+      break;
+    }
+    if (data.source == TraceReader::SOURCE_FILE) {
+      binary_file_names.insert(make_pair(move(data.file_name), km.fsname()));
+    }
+  }
+
+  return sources(binary_file_names, comp_dir_substitutions, false);
+}
+
+int ExplicitSourcesCommand::run(vector<string>& args) {
+  map<string, string> comp_dir_substitutions;
+  while (parse_sources_option(args, comp_dir_substitutions)) {
+  }
+
+  // (Trace file name, original file name) pairs
+  map<string, string> binary_file_names;
+  for (auto arg : args) {
+    struct stat statbuf;
+    int ret = stat(arg.c_str(), &statbuf);
+    if (ret < 0) {
+      FATAL() << "Failed to stat `" << arg << "`";
+    }
+    if (!S_ISREG(statbuf.st_mode)) {
+      continue;
+    }
+
+    ScopedFd fd = ScopedFd(arg.c_str(), O_RDONLY, 0);
+    if (!fd.is_open()) {
+      LOG(error) << "Failed to open `" << arg << "`";
+      return 1;
+    }
+
+    ElfFileReader reader(fd);
+    auto buildid = reader.read_buildid();
+    if (buildid.empty()) {
+      LOG(warn) << "No build-id for `" << arg << "`";
+      continue;
+    }
+    binary_file_names.insert(make_pair(move(buildid), arg));
+  }
+
+  return sources(binary_file_names, comp_dir_substitutions, true);
 }
 
 } // namespace rr
