@@ -24,6 +24,7 @@
 #include "RecordTask.h"
 #include "TaskishUid.h"
 #include "core.h"
+#include "kernel_abi.h"
 #include "kernel_supplement.h"
 #include "log.h"
 #include "rr_trace.capnp.h"
@@ -189,8 +190,9 @@ private:
 };
 
 TraceStream::TraceStream(const string& trace_dir, FrameTime initial_time)
-    : trace_dir(real_path(trace_dir))
-    , global_time(initial_time)
+    : trace_dir(real_path(trace_dir)),
+      bind_to_cpu(-1),
+      global_time(initial_time)
    {}
 
 string TraceStream::file_data_clone_file_name(const TaskUid& tuid) {
@@ -241,18 +243,26 @@ static trace::Arch to_trace_arch(SupportedArch arch) {
       return trace::Arch::X86;
     case x86_64:
       return trace::Arch::X8664;
+    case aarch64:
+      return trace::Arch::AARCH64;
     default:
       FATAL() << "Unknown arch";
       return trace::Arch::X86;
   }
 }
 
+static trace::CpuTriState to_tristate(bool value) {
+  return value ? trace::CpuTriState::KNOWN_TRUE : trace::CpuTriState::KNOWN_FALSE;
+}
+
 static SupportedArch from_trace_arch(trace::Arch arch) {
   switch (arch) {
     case trace::Arch::X86:
       return x86;
     case trace::Arch::X8664:
       return x86_64;
+    case trace::Arch::AARCH64:
+      return aarch64;
     default:
       FATAL() << "Unknown arch";
       return x86;
@@ -412,7 +422,7 @@ void TraceWriter::write_frame(RecordTask* t, const Event& ev,
   frame.setArch(to_trace_arch(t->arch()));
   if (registers) {
     // Avoid dynamic allocation and copy
-    auto raw_regs = registers->get_ptrace_for_self_arch();
+    auto raw_regs = registers->get_regs_for_trace();
     frame.initRegisters().setRaw(Data::Reader(raw_regs.data, raw_regs.size));
   }
   if (extra_registers) {
@@ -426,7 +436,13 @@ void TraceWriter::write_frame(RecordTask* t, const Event& ev,
       event.setInstructionTrap(Void());
       break;
     case EV_PATCH_SYSCALL:
-      event.setPatchSyscall(Void());
+      if (ev.PatchSyscall().patch_vsyscall) {
+        event.setPatchVsyscall(Void());
+      } else if (ev.PatchSyscall().patch_after_syscall) {
+        event.setPatchAfterSyscall(Void());
+      } else {
+        event.setPatchSyscall(Void());
+      }
       break;
     case EV_SYSCALLBUF_ABORT_COMMIT:
       event.setSyscallbufAbortCommit(Void());
@@ -533,16 +549,29 @@ TraceFrame TraceReader::read_frame() {
   ret.recorded_regs.set_arch(arch);
   auto reg_data = frame.getRegisters().getRaw();
   if (reg_data.size()) {
-    ret.recorded_regs.set_from_ptrace_for_arch(arch, reg_data.begin(),
-                                               reg_data.size());
+    ret.recorded_regs.set_from_trace(arch, reg_data.begin(),
+                                     reg_data.size());
   }
   auto extra_reg_data = frame.getExtraRegisters().getRaw();
   if (extra_reg_data.size()) {
+    ExtraRegisters::Format fmt;
+    switch (arch) {
+      default:
+        FATAL() << "Unknown architecture";
+        RR_FALLTHROUGH;
+      case x86:
+      case x86_64:
+        fmt = ExtraRegisters::XSAVE;
+        break;
+      case aarch64:
+        fmt = ExtraRegisters::NT_FPR;
+        break;
+    }
     bool ok = ret.recorded_extra_regs.set_to_raw_data(
-        arch, ExtraRegisters::XSAVE, extra_reg_data.begin(),
+        arch, fmt, extra_reg_data.begin(),
         extra_reg_data.size(), xsave_layout_from_trace(cpuid_records()));
     if (!ok) {
-      FATAL() << "Invalid XSAVE data in trace";
+      FATAL() << "Invalid extended register data in trace";
     }
   } else {
     ret.recorded_extra_regs = ExtraRegisters(arch);
@@ -556,6 +585,14 @@ TraceFrame TraceReader::read_frame() {
     case trace::Frame::Event::PATCH_SYSCALL:
       ret.ev = Event::patch_syscall();
       break;
+    case trace::Frame::Event::PATCH_VSYSCALL:
+      ret.ev = Event::patch_syscall();
+      ret.ev.PatchSyscall().patch_vsyscall = true;
+      break;
+    case trace::Frame::Event::PATCH_AFTER_SYSCALL:
+      ret.ev = Event::patch_syscall();
+      ret.ev.PatchSyscall().patch_after_syscall = true;
+      break;
     case trace::Frame::Event::SYSCALLBUF_ABORT_COMMIT:
       ret.ev = Event::syscallbuf_abort_commit();
       break;
@@ -670,6 +707,9 @@ void TraceWriter::write_task_event(const TraceTaskEvent& event) {
     case TraceTaskEvent::EXIT:
       task.initExit().setExitStatus(event.exit_status().get());
       break;
+    case TraceTaskEvent::DETACH:
+      task.initDetach();
+      break;
     case TraceTaskEvent::NONE:
       DEBUG_ASSERT(0 && "Writing NONE TraceTaskEvent");
       break;
@@ -703,7 +743,10 @@ TraceTaskEvent TraceReader::read_task_event(FrameTime* time) {
       r.type_ = TraceTaskEvent::CLONE;
       auto clone = task.getClone();
       r.parent_tid_ = i32_to_tid(clone.getParentTid());
-      r.own_ns_tid_ = i32_to_tid(clone.getOwnNsTid());
+      // The record task could have died before we were able to compute this,
+      // which we handle fine (since we'll see the exit event right away),
+      // but we can't use i32_to_tid, which will assert if the tid is invalid.
+      r.own_ns_tid_ = (pid_t)clone.getOwnNsTid();
       r.clone_flags_ = clone.getFlags();
       LOG(debug) << "Reading event for " << task.getFrameTime()
                  << ": parent=" << r.parent_tid_ << " tid=" << r.tid_;
@@ -725,6 +768,9 @@ TraceTaskEvent TraceReader::read_task_event(FrameTime* time) {
       r.type_ = TraceTaskEvent::EXIT;
       r.exit_status_ = WaitStatus(task.getExit().getExitStatus());
       break;
+    case trace::TaskEvent::Which::DETACH:
+      r.type_ = TraceTaskEvent::DETACH;
+      break;
     default:
       DEBUG_ASSERT(0 && "Unknown TraceEvent type");
       break;
@@ -738,14 +784,15 @@ static string base_file_name(const string& file_name) {
                                         : file_name;
 }
 
-bool TraceWriter::try_hardlink_file(const string& file_name,
+bool TraceWriter::try_hardlink_file(const string& real_file_name,
+                                    const string& access_file_name,
                                     std::string* new_name) {
   char count_str[20];
   sprintf(count_str, "%d", mmap_count);
 
   string path =
-      string("mmap_hardlink_") + count_str + "_" + base_file_name(file_name);
-  int ret = link(file_name.c_str(), (dir() + "/" + path).c_str());
+      string("mmap_hardlink_") + count_str + "_" + base_file_name(real_file_name);
+  int ret = linkat(AT_FDCWD, access_file_name.c_str(), AT_FDCWD, (dir() + "/" + path).c_str(), AT_SYMLINK_FOLLOW);
   if (ret < 0) {
     return false;
   }
@@ -753,8 +800,9 @@ bool TraceWriter::try_hardlink_file(const string& file_name,
   return true;
 }
 
-bool TraceWriter::try_clone_file(RecordTask* t, const string& file_name,
-                                 string* new_name) {
+bool TraceWriter::try_clone_file(RecordTask* t,
+    const string& real_file_name, const string& access_file_name,
+    string* new_name) {
   if (!t->session().use_file_cloning()) {
     return false;
   }
@@ -763,9 +811,9 @@ bool TraceWriter::try_clone_file(RecordTask* t, const string& file_name,
   sprintf(count_str, "%d", mmap_count);
 
   string path =
-      string("mmap_clone_") + count_str + "_" + base_file_name(file_name);
+      string("mmap_clone_") + count_str + "_" + base_file_name(real_file_name);
 
-  ScopedFd src(file_name.c_str(), O_RDONLY);
+  ScopedFd src(access_file_name.c_str(), O_RDONLY);
   if (!src.is_open()) {
     return false;
   }
@@ -786,17 +834,18 @@ bool TraceWriter::try_clone_file(RecordTask* t, const string& file_name,
   return true;
 }
 
-bool TraceWriter::copy_file(const std::string& file_name,
+bool TraceWriter::copy_file(const std::string& real_file_name,
+                            const std::string& access_file_name,
                             std::string* new_name) {
   char count_str[20];
   sprintf(count_str, "%d", mmap_count);
 
   string path =
-      string("mmap_copy_") + count_str + "_" + base_file_name(file_name);
+      string("mmap_copy_") + count_str + "_" + base_file_name(real_file_name);
 
-  ScopedFd src(file_name.c_str(), O_RDONLY);
+  ScopedFd src(access_file_name.c_str(), O_RDONLY);
   if (!src.is_open()) {
-    LOG(debug) << "Can't open " << file_name;
+    LOG(debug) << "Can't open " << access_file_name;
     return false;
   }
   string dest_path = dir() + "/" + path;
@@ -810,36 +859,13 @@ bool TraceWriter::copy_file(const std::string& file_name,
   return rr::copy_file(dest, src);
 }
 
-/**
- * Given `file_name`, where `file_name` is relative to our root directory
- * but is in the mount namespace of `t`, try to make it a file we can read.
- */
-static string try_make_process_file_name(RecordTask* t,
-                                         const std::string& file_name) {
-  char proc_root[32];
-  // /proc/<pid>/root has magical properties; not only is it a link, but
-  // it links to a view of the filesystem as the process sees it, taking into
-  // account the process mount namespace etc.
-  snprintf(proc_root, sizeof(proc_root), "/proc/%d/root", t->tid);
-  char root[PATH_MAX];
-  ssize_t ret = readlink(proc_root, root, sizeof(root) - 1);
-  ASSERT(t, ret >= 0);
-  root[ret] = 0;
-
-  if (strncmp(root, file_name.c_str(), ret)) {
-    LOG(debug) << "File " << file_name << " is outside known root "
-               << proc_root;
-    return file_name;
-  }
-  return string(proc_root) + (ret == 1 ? file_name : file_name.substr(ret));
-}
-
 static bool starts_with(const string& s, const string& with) {
   return s.find(with) == 0;
 }
 
 TraceWriter::RecordInTrace TraceWriter::write_mapped_region(
-    RecordTask* t, const KernelMapping& km, const struct stat& stat,
+    RecordTask* t, const KernelMapping& km,
+    const struct stat& stat, const std::string &file_name,
     const vector<TraceRemoteFd>& extra_fds, MappingOrigin origin,
     bool skip_monitoring_mapped_fd) {
   MallocMessageBuilder map_msg;
@@ -869,6 +895,27 @@ TraceWriter::RecordInTrace TraceWriter::write_mapped_region(
   auto src = map.getSource();
   string backing_file_name;
 
+  auto copy_file_or_trace = [&](string file_name) {
+    // Make executable files accessible to debuggers by copying the whole
+    // thing into the trace directory. We don't get to compress the data and
+    // the entire file is copied, not just the used region, which is why we
+    // don't do this for all files.
+    // Don't bother trying to copy [vdso].
+    // Don't try to copy files that use shared mappings. We do not want to
+    // create a shared mapping of a file stored in the trace. This means
+    // debuggers can't find the file, but the Linux loader doesn't create
+    // shared mappings so situations where a shared-mapped executable contains
+    // usable debug info should be very rare at best...
+    string backing_file_name;
+    if ((km.prot() & PROT_EXEC) &&
+        copy_file(km.fsname(), file_name, &backing_file_name) &&
+        !(km.flags() & MAP_SHARED)) {
+      src.initFile().setBackingFileName(str_to_data(backing_file_name));
+    } else {
+      src.setTrace();
+    }
+  };
+
   if (origin == REMAP_MAPPING || origin == PATCH_MAPPING ||
       origin == RR_BUFFER_MAPPING) {
     src.setZero();
@@ -880,52 +927,54 @@ TraceWriter::RecordInTrace TraceWriter::write_mapped_region(
   } else if (!starts_with(km.fsname(), "/")) {
     src.setTrace();
   } else {
-    string file_name = try_make_process_file_name(t, km.fsname());
     auto assumed_immutable =
         files_assumed_immutable.find(make_pair(stat.st_dev, stat.st_ino));
 
     if (assumed_immutable != files_assumed_immutable.end()) {
       src.initFile().setBackingFileName(str_to_data(assumed_immutable->second));
     } else if ((km.flags() & MAP_PRIVATE) &&
-               try_clone_file(t, file_name, &backing_file_name)) {
+               try_clone_file(t, km.fsname(), file_name, &backing_file_name)) {
       src.initFile().setBackingFileName(str_to_data(backing_file_name));
-    } else if (should_copy_mmap_region(km, stat)) {
-      // Make executable files accessible to debuggers by copying the whole
-      // thing into the trace directory. We don't get to compress the data and
-      // the entire file is copied, not just the used region, which is why we
-      // don't do this for all files.
-      // Don't bother trying to copy [vdso].
-      // Don't try to copy files that use shared mappings. We do not want to
-      // create a shared mapping of a file stored in the trace. This means
-      // debuggers can't find the file, but the Linux loader doesn't create
-      // shared mappings so situations where a shared-mapped executable contains
-      // usable debug info should be very rare at best...
-      if ((km.prot() & PROT_EXEC) &&
-          copy_file(file_name, &backing_file_name) &&
-          !(km.flags() & MAP_SHARED)) {
-        src.initFile().setBackingFileName(str_to_data(backing_file_name));
-      } else {
-        src.setTrace();
-      }
+    } else if (should_copy_mmap_region(km, file_name, stat)) {
+      copy_file_or_trace(file_name);
     } else {
       // should_copy_mmap_region's heuristics determined it was OK to just map
       // the file here even if it's MAP_SHARED. So try cloning again to avoid
       // the possibility of the file changing between recording and replay.
-      if (!try_clone_file(t, file_name, &backing_file_name)) {
+      if (try_clone_file(t, km.fsname(), file_name, &backing_file_name)) {
+        src.initFile().setBackingFileName(str_to_data(backing_file_name));
+      } else {
         // Try hardlinking file into the trace directory. This will avoid
         // replay failures if the original file is deleted or replaced (but not
         // if it is overwritten in-place). If try_hardlink_file fails it
         // just returns the original file name.
         // A relative backing_file_name is relative to the trace directory.
-        if (!try_hardlink_file(file_name, &backing_file_name)) {
+        if (!try_hardlink_file(km.fsname(), file_name, &backing_file_name)) {
           // Don't ever use `file_name` for the `backing_file_name` because it
           // contains the pid of a recorded process and will not work!
           backing_file_name = km.fsname();
-        }
-        files_assumed_immutable.insert(
+          // If the backing_file_name refers to a different file (e.g.
+          // because we have a file in our mount namespace that matches the
+          // fsname in the Task's mount namespace), then copy the file even
+          // if the heuristics above thought we could get away with leaving it
+          // in place. Otherwise, we're essentially guaranteed to have a broken
+          // recording.
+          struct stat backing_stat;
+          int err = ::stat(backing_file_name.c_str(), &backing_stat);
+          if (err != 0 || backing_stat.st_dev != km.device() ||
+              backing_stat.st_ino != km.inode()) {
+            copy_file_or_trace(file_name);
+          } else {
+            files_assumed_immutable.insert(
+                make_pair(make_pair(stat.st_dev, stat.st_ino), backing_file_name));
+            src.initFile().setBackingFileName(str_to_data(backing_file_name));
+          }
+        } else {
+          files_assumed_immutable.insert(
             make_pair(make_pair(stat.st_dev, stat.st_ino), backing_file_name));
+          src.initFile().setBackingFileName(str_to_data(backing_file_name));
+        }
       }
-      src.initFile().setBackingFileName(str_to_data(backing_file_name));
     }
   }
 
@@ -1184,7 +1233,6 @@ static string make_trace_dir(const string& exe_path, const string& output_trace_
 #define STR(x) STR_HELPER(x)
 
 TraceWriter::TraceWriter(const std::string& file_name,
-                         int bind_to_cpu,
                          const string& output_trace_dir,
                          TicksSemantics ticks_semantics_)
     : TraceStream(make_trace_dir(file_name, output_trace_dir),
@@ -1193,9 +1241,11 @@ TraceWriter::TraceWriter(const std::string& file_name,
                   1),
       ticks_semantics_(ticks_semantics_),
       mmap_count(0),
-      has_cpuid_faulting_(false) {
+      has_cpuid_faulting_(false),
+      xsave_fip_fdp_quirk_(false),
+      fdp_exception_only_quirk_(false),
+      clear_fip_fdp_(false) {
   this->ticks_semantics_ = ticks_semantics_;
-  this->bind_to_cpu = bind_to_cpu;
 
   for (Substream s = SUBSTREAM_FIRST; s < SUBSTREAM_COUNT; ++s) {
     writers[s] = unique_ptr<CompressedWriter>(new CompressedWriter(
@@ -1265,15 +1315,27 @@ void TraceWriter::close(CloseStatus status, const TraceUuid* uuid) {
   MallocMessageBuilder header_msg;
   trace::Header::Builder header = header_msg.initRoot<trace::Header>();
   header.setBindToCpu(this->bind_to_cpu);
-  header.setHasCpuidFaulting(has_cpuid_faulting_);
-  header.setCpuidRecords(
-      Data::Reader(reinterpret_cast<const uint8_t*>(cpuid_records.data()),
-                   cpuid_records.size() * sizeof(CPUIDRecord)));
-  header.setXcr0(xcr0());
   header.setTicksSemantics(
     to_trace_ticks_semantics(PerfCounters::default_ticks_semantics()));
   header.setSyscallbufProtocolVersion(SYSCALLBUF_PROTOCOL_VERSION);
   header.setPreloadThreadLocalsRecorded(true);
+  header.setRrcallBase(syscall_number_for_rrcall_init_preload(x86_64));
+
+  header.setNativeArch(to_trace_arch(NativeArch::arch()));
+  if (NativeArch::is_x86ish())
+  {
+    auto x86data = header.initX86();
+    x86data.setHasCpuidFaulting(has_cpuid_faulting_);
+    x86data.setCpuidRecords(
+        Data::Reader(reinterpret_cast<const uint8_t*>(cpuid_records.data()),
+                    cpuid_records.size() * sizeof(CPUIDRecord)));
+    x86data.setXcr0(xcr0());
+    x86data.setXsaveFipFdpQuirk(to_tristate(xsave_fip_fdp_quirk_));
+    x86data.setFdpExceptionOnlyQuirk(to_tristate(fdp_exception_only_quirk_));
+    x86data.setClearFipFdp(clear_fip_fdp_);
+  }
+
+  header.setExplicitProcMem(false);
   // Add a random UUID to the trace metadata. This lets tools identify a trace
   // easily.
   if (!uuid) {
@@ -1324,12 +1386,14 @@ TraceFrame TraceReader::peek_frame() {
   auto& events = reader(EVENTS);
   events.save_state();
   auto saved_time = global_time;
+  auto saved_raw_recs = raw_recs;
   TraceFrame frame;
   if (!at_end()) {
     frame = read_frame();
   }
   events.restore_state();
   global_time = saved_time;
+  raw_recs = saved_raw_recs;
   return frame;
 }
 
@@ -1410,18 +1474,10 @@ TraceReader::TraceReader(const string& dir)
 
   trace::Header::Reader header = header_msg.getRoot<trace::Header>();
   bind_to_cpu = header.getBindToCpu();
-  trace_uses_cpuid_faulting = header.getHasCpuidFaulting();
-  Data::Reader cpuid_records_bytes = header.getCpuidRecords();
-  size_t len = cpuid_records_bytes.size() / sizeof(CPUIDRecord);
-  if (cpuid_records_bytes.size() != len * sizeof(CPUIDRecord)) {
-    FATAL() << "Invalid CPUID records length";
-  }
-  cpuid_records_.resize(len);
-  memcpy(cpuid_records_.data(), cpuid_records_bytes.begin(),
-         len * sizeof(CPUIDRecord));
-  xcr0_ = header.getXcr0();
   preload_thread_locals_recorded_ = header.getPreloadThreadLocalsRecorded();
   ticks_semantics_ = from_trace_ticks_semantics(header.getTicksSemantics());
+  rrcall_base_ = header.getRrcallBase();
+  explicit_proc_mem_ = header.getExplicitProcMem();
   Data::Reader uuid = header.getUuid();
   uuid_ = unique_ptr<TraceUuid>(new TraceUuid());
   if (uuid.size() != sizeof(uuid_->bytes)) {
@@ -1429,6 +1485,22 @@ TraceReader::TraceReader(const string& dir)
   }
   memcpy(uuid_->bytes, uuid.begin(), sizeof(uuid_->bytes));
 
+  arch_ = from_trace_arch(header.getNativeArch());
+  if (is_x86ish(arch_)) {
+    auto x86data = header.getX86();
+    trace_uses_cpuid_faulting = x86data.getHasCpuidFaulting();
+    Data::Reader cpuid_records_bytes = x86data.getCpuidRecords();
+    size_t len = cpuid_records_bytes.size() / sizeof(CPUIDRecord);
+    if (cpuid_records_bytes.size() != len * sizeof(CPUIDRecord)) {
+      FATAL() << "Invalid CPUID records length";
+    }
+    cpuid_records_.resize(len);
+    memcpy(cpuid_records_.data(), cpuid_records_bytes.begin(),
+          len * sizeof(CPUIDRecord));
+    xcr0_ = x86data.getXcr0();
+    clear_fip_fdp_ = x86data.getClearFipFdp();
+  }
+
   // Set the global time at 0, so that when we tick it for the first
   // event, it matches the initial global time at recording, 1.
   global_time = 0;
@@ -1452,6 +1524,10 @@ TraceReader::TraceReader(const TraceReader& other)
   raw_recs = other.raw_recs;
   xcr0_ = other.xcr0_;
   preload_thread_locals_recorded_ = other.preload_thread_locals_recorded_;
+  rrcall_base_ = other.rrcall_base_;
+  arch_ = other.arch_;
+  explicit_proc_mem_ = other.explicit_proc_mem_;
+  clear_fip_fdp_ = other.clear_fip_fdp_;
 }
 
 TraceReader::~TraceReader() {}
