@@ -198,6 +198,7 @@ public:
    */
   RecordInTrace write_mapped_region(RecordTask* t, const KernelMapping& map,
                                     const struct stat& stat,
+                                    const std::string &file_name,
                                     const std::vector<TraceRemoteFd>& extra_fds,
                                     MappingOrigin origin = SYSCALL_MAPPING,
                                     bool skip_monitoring_mapped_fd = false);
@@ -231,8 +232,8 @@ public:
    * The trace name is determined by |file_name| and _RR_TRACE_DIR (if set)
    * or by setting -o=<OUTPUT_TRACE_DIR>.
    */
-  TraceWriter(const std::string& file_name, int bind_to_cpu,
-              const string& output_trace_dir, TicksSemantics ticks_semantics_);
+  TraceWriter(const std::string& file_name,
+              const string& output_trace_dir, TicksSemantics ticks_semantics);
 
   /**
    * Called after the calling thread is actually bound to |bind_to_cpu|.
@@ -240,6 +241,11 @@ public:
   void setup_cpuid_records(bool has_cpuid_faulting,
                            const DisableCPUIDFeatures& disable_cpuid_features);
 
+  void set_xsave_fip_fdp_quirk(bool value) { xsave_fip_fdp_quirk_ = value; }
+  void set_fdp_exception_only_quirk(bool value) { fdp_exception_only_quirk_ = value; }
+  void set_clear_fip_fdp(bool value) { clear_fip_fdp_ = value; }
+  bool clear_fip_fdp() const { return clear_fip_fdp_; }
+
   enum CloseStatus {
     /**
      * Trace completed normally and can be replayed.
@@ -266,10 +272,13 @@ public:
   TicksSemantics ticks_semantics() const { return ticks_semantics_; }
 
 private:
-  bool try_hardlink_file(const std::string& file_name, std::string* new_name);
-  bool try_clone_file(RecordTask* t, const std::string& file_name,
+  bool try_hardlink_file(const std::string& real_file_name,
+                         const std::string& access_file_name, std::string* new_name);
+  bool try_clone_file(RecordTask* t, const std::string& real_file_name,
+                      const std::string& access_file_name,
                       std::string* new_name);
-  bool copy_file(const std::string& file_name, std::string* new_name);
+  bool copy_file(const std::string& real_file_name,
+                 const std::string& access_file_name, std::string* new_name);
 
   CompressedWriter& writer(Substream s) { return *writers[s]; }
   const CompressedWriter& writer(Substream s) const { return *writers[s]; }
@@ -291,6 +300,9 @@ private:
   ScopedFd version_fd;
   uint32_t mmap_count;
   bool has_cpuid_faulting_;
+  bool xsave_fip_fdp_quirk_;
+  bool fdp_exception_only_quirk_;
+  bool clear_fip_fdp_;
   bool supports_file_data_cloning_;
 };
 
@@ -401,6 +413,7 @@ public:
   }
   bool uses_cpuid_faulting() const { return trace_uses_cpuid_faulting; }
   uint64_t xcr0() const;
+  bool clear_fip_fdp() const { return clear_fip_fdp_; }
   // Prior to issue 2370, we did not emit mapping into the trace for the
   // preload_thread_locals mapping if it was created by a clone(2) without
   // CLONE_VM. This is true if that has been fixed.
@@ -411,6 +424,14 @@ public:
 
   double recording_time() const { return monotonic_time_; }
 
+  // The base syscall number for rr syscalls in this trace
+  int rrcall_base() const { return rrcall_base_; }
+
+  SupportedArch arch() const { return arch_; }
+
+  // Whether the /proc/<pid>/mem calls were explicitly recorded in this trace
+  bool explicit_proc_mem() const { return explicit_proc_mem_; }
+
 private:
   CompressedReader& reader(Substream s) { return *readers[s]; }
   const CompressedReader& reader(Substream s) const { return *readers[s]; }
@@ -424,6 +445,10 @@ private:
   std::unique_ptr<TraceUuid> uuid_;
   bool trace_uses_cpuid_faulting;
   bool preload_thread_locals_recorded_;
+  bool clear_fip_fdp_;
+  int rrcall_base_;
+  SupportedArch arch_;
+  bool explicit_proc_mem_;
 };
 
 extern std::string trace_save_dir();
