@@ -12,6 +12,11 @@
 
 namespace rr {
 
+/**
+ * Return the symbolic name of the architecture `arch`.
+ */
+std::string arch_name(SupportedArch arch);
+
 /**
  * Return the symbolic name of |syscall|, f.e. "read", or "syscall(%d)"
  * if unknown.
@@ -28,6 +33,7 @@ std::string ptrace_event_name(int event);
  * Return the symbolic name of the PTRACE_ |request|, or "PTRACE_REQUEST(%d)" if
  * unknown.
  */
+template <typename Arch>
 std::string ptrace_req_name(int request);
 
 /**
@@ -63,6 +69,11 @@ int shm_flags_to_mmap_prot(int flags);
  */
 std::string xsave_feature_string(uint64_t xsave_features);
 
+/**
+ * Return whether this is a core dumping signal or not.
+ */
+bool is_coredumping_signal(int signo);
+
 } // namespace rr
 
 #endif
