@@ -69,29 +69,11 @@ void print_global_options(FILE* out) {
       "                             'Ivy Bridge'. Note that rr will not work "
       "with\n"
       "                             Intel Merom or Penryn microarchitectures.\n"
-      "  -C, --checksum={on-syscalls,on-all-events}|FROM_TIME\n"
-      "                             compute and store (during recording) or\n"
-      "                             read and verify (during replay) checksums\n"
-      "                             of each of a tracee's memory mappings "
-      "either\n"
-      "                             at the end of all syscalls "
-      "(`on-syscalls'),\n"
-      "                             at all events (`on-all-events'), or \n"
-      "                             starting from a global timepoint "
-      "FROM_TIME\n"
-      "  -D, --dump-on=<SYSCALL_NUM|-SIGNAL_NUM>\n"
-      "                             dump memory at SYSCALL or SIGNAL to the\n"
-      "                             file "
-      "`[trace_dir]/[tid].[time]_{rec,rep}':\n"
-      "                             `_rec' for dumps during recording, `_rep'\n"
-      "                             for dumps during replay\n"
       "  -F, --force-things         force rr to do some things that don't "
       "seem\n"
       "                             like good ideas, for example launching an\n"
       "                             interactive emergency debugger if stderr\n"
       "                             isn't a tty.\n"
-      "  -K, --check-cached-mmaps   verify that cached task mmaps match "
-      "/proc/maps\n"
       "  -E, --fatal-errors         any warning or error that is printed is\n"
       "                             treated as fatal\n"
       "  -M, --mark-stdio           mark stdio writes with [rr <PID> <EV>]\n"
@@ -102,7 +84,6 @@ void print_global_options(FILE* out) {
       "  -S, --suppress-environment-warnings\n"
       "                             suppress warnings about issues in the\n"
       "                             environment that rr has no control over\n"
-      "  -T, --dump-at=TIME         dump memory at global timepoint TIME\n"
       "\n"
       "Use RR_LOG to control logging; e.g. RR_LOG=all:warn,Task:debug\n",
       out);
