@@ -12,7 +12,10 @@
 #include <sys/resource.h>
 #include <sys/user.h>
 #include <syscall.h>
+
+#if defined(__i386__) || defined(__x86_64__)
 #include <x86intrin.h>
+#endif
 
 #include "preload/preload_interface.h"
 
@@ -32,7 +35,14 @@ using namespace std;
 
 namespace rr {
 
-static __inline__ unsigned long long rdtsc(void) { return __rdtsc(); }
+static __inline__ unsigned long long rdtsc(void) {
+#if defined(__i386__) || defined(__x86_64__)
+  return __rdtsc();
+#else
+  FATAL() << "Reached x86-only code path on non-x86 architecture";
+  return 0;
+#endif
+}
 
 static void restore_sighandler_if_not_default(RecordTask* t, int sig) {
   if (t->sig_disposition(sig) != SIGNAL_DEFAULT) {
@@ -156,9 +166,14 @@ static bool try_grow_map(RecordTask* t, siginfo_t* si) {
     LOG(debug) << "try_grow_map " << addr << ": address would be in guard page";
     return false;
   }
-  struct rlimit stack_limit;
   remote_ptr<void> limit_bottom;
+#if defined (__i386__)
+  struct rlimit stack_limit;
   int ret = prlimit(t->tid, RLIMIT_STACK, NULL, &stack_limit);
+#else
+  struct rlimit64 stack_limit;
+  int ret = syscall(__NR_prlimit64, t->tid, RLIMIT_STACK, NULL, &stack_limit);
+#endif
   if (ret >= 0 && stack_limit.rlim_cur != RLIM_INFINITY) {
     limit_bottom = ceil_page_size(it->map.end() - stack_limit.rlim_cur);
     if (limit_bottom > addr) {
@@ -197,7 +212,7 @@ static bool try_grow_map(RecordTask* t, siginfo_t* si) {
       t->vm()->map(t, new_start, it->map.start() - new_start, it->map.prot(),
                    it->map.flags() | MAP_ANONYMOUS, 0, string(),
                    KernelMapping::NO_DEVICE, KernelMapping::NO_INODE);
-  t->trace_writer().write_mapped_region(t, km, km.fake_stat(), vector<TraceRemoteFd>());
+  t->trace_writer().write_mapped_region(t, km, km.fake_stat(), km.fsname(), vector<TraceRemoteFd>());
   // No need to flush syscallbuf here. It's safe to map these pages "early"
   // before they're really needed.
   t->record_event(Event::grow_map(), RecordTask::DONT_FLUSH_SYSCALLBUF);
@@ -260,7 +275,7 @@ bool handle_syscallbuf_breakpoint(RecordTask* t) {
   }
 
   Registers r = t->regs();
-  r.set_ip(r.ip().decrement_by_bkpt_insn_length(t->arch()));
+  r.set_ip(r.ip().undo_executed_bkpt(t->arch()));
   t->set_regs(r);
 
   if (t->is_at_traced_syscall_entry()) {
