@@ -1,6 +1,6 @@
 /* -*- Mode: C; tab-width: 8; c-basic-offset: 2; indent-tabs-mode: nil; -*- */
 
-#include "rrutil.h"
+#include "util.h"
 
 int main(__attribute__((unused)) int argc, char* argv[]) {
   int fd = open(argv[0], O_RDONLY);
