@@ -4,7 +4,7 @@
 
 static void sighandler(int sig, siginfo_t* si,
                        __attribute__((unused)) void* utp) {
-  test_assert(SIGSEGV == sig && si->si_addr == (void*)0x42);
+  test_assert(SIGSEGV == sig && si->si_addr == (void*)0x44);
 
   atomic_puts("EXIT-SUCCESS");
   _exit(0);
@@ -18,6 +18,6 @@ int main(void) {
   act.sa_flags = SA_SIGINFO;
   sigaction(SIGSEGV, &act, NULL);
 
-  __asm__ __volatile__("call *0x42");
+  ((void (*)(void))(0x44))();
   return 0;
 }
