@@ -1,6 +1,6 @@
 /* -*- Mode: C; tab-width: 8; c-basic-offset: 2; indent-tabs-mode: nil; -*- */
 
-#include "rrutil.h"
+#include "util.h"
 
 static int futex(int* uaddr, int op, int val, const struct timespec* timeout,
                  int* uaddr2, int val2) {
