@@ -29,7 +29,7 @@ const char* sockaddr_hw_name(const struct sockaddr* addr) {
 static void get_ifconfig(int sockfd, struct ifreq* req, struct ifreq* eth_req) {
   struct {
     struct ifreq ifaces[100];
-  } * ifaces;
+  }* ifaces;
   struct ifconf* ifconf;
   int ret;
   ssize_t num_ifaces;
@@ -133,8 +133,9 @@ static int generic_wireless_request_by_name_internal(int sockfd,
   atomic_printf("%s(ret:%d): %s: ", nr_str, ret, (*wreq)->ifr_name);
   if (-1 == ret) {
     atomic_printf("WARNING: %s doesn't support ioctl %s\n", name, nr_str);
-    /* "bond" network devices can return ENODEV. */
-    test_assert(EOPNOTSUPP == err || EPERM == err || EINVAL == err || ENODEV == err);
+    /* "bond" network devices can return ENODEV.
+       Some virtual ethernet devices can return ENOTTY. */
+    test_assert(EOPNOTSUPP == err || EPERM == err || EINVAL == err || ENODEV == err || ENOTTY == err);
   }
   return ret;
 }
@@ -156,13 +157,219 @@ static int generic_wireless_request_by_name_internal(int sockfd,
                 wreq->u.name.value, wreq->u.name.fixed ? "yes" : "no",         \
                 wreq->u.name.disabled ? "yes" : "no", wreq->u.name.flags));
 
+#define GENERIC_ETHTOOL_REQUEST_BY_NAME(ptr, nr) \
+  (ptr)->cmd = nr; \
+  req->ifr_data = (char*)ptr; \
+  ret = ioctl(sockfd, SIOCETHTOOL, req); \
+  VERIFY_GUARD(req); \
+  VERIFY_GUARD(ptr); \
+  err = errno; \
+  atomic_printf("SIOCETHTOOL(ret:%d) " #nr ": %s ethtool data: ", ret, req->ifr_name); \
+  if (-1 == ret) { \
+    atomic_printf("WARNING: %s doesn't appear to support SIOCETHTOOL " #nr "\n", req->ifr_name); \
+    test_assert(EOPNOTSUPP == err || EPERM == err); \
+  }
+
+static void ethtool(int sockfd, struct ifreq* req) {
+  struct ethtool_cmd* et_set;
+  struct ethtool_drvinfo* et_drvinfo;
+  struct ethtool_wolinfo* et_wolinfo;
+  struct {
+    struct ethtool_regs et;
+    uint8_t data[32];
+  }* et_regs;
+  struct {
+    struct ethtool_eeprom et;
+    uint8_t data[32];
+  }* et_eeprom;
+  struct ethtool_eee* et_eee;
+  struct ethtool_modinfo* et_modinfo;
+  struct ethtool_coalesce* et_coalesce;
+  struct ethtool_ringparam* et_ringparam;
+  struct ethtool_channels* et_channels;
+  struct ethtool_pauseparam* et_pauseparam;
+  struct {
+    struct ethtool_sset_info et;
+    uint32_t data[8];
+  }* et_sset_info;
+  struct {
+    struct ethtool_gfeatures et;
+    struct ethtool_get_features_block features[20];
+  }* et_gfeatures;
+  struct {
+    struct ethtool_perm_addr et;
+    uint8_t data[32];
+  }* et_perm_addr;
+  int i;
+  int err;
+  int ret;
+
+  ALLOCATE_GUARD(et_set, 'b');
+  GENERIC_ETHTOOL_REQUEST_BY_NAME(et_set, ETHTOOL_GSET);
+  if (-1 == ret) {
+    return;
+  }
+  atomic_printf("speed:%#x duplex:%#x port:%#x physaddr:%#x, maxtxpkt:%u "
+                "maxrxpkt:%u ...\n",
+                ethtool_cmd_speed(et_set), et_set->duplex, et_set->port,
+                et_set->phy_address, et_set->maxtxpkt, et_set->maxrxpkt);
+
+  ALLOCATE_GUARD(et_drvinfo, 'c');
+  GENERIC_ETHTOOL_REQUEST_BY_NAME(et_drvinfo, ETHTOOL_GDRVINFO);
+  if (-1 != ret) {
+    #ifdef ETHTOOL_EROMVERS_LEN
+    atomic_printf("driver:%s version:%s fw_version:%s bus_info:%s erom_version:%s\n",
+                  et_drvinfo->driver, et_drvinfo->version, et_drvinfo->fw_version,
+                  et_drvinfo->bus_info, et_drvinfo->erom_version);
+    #else
+    atomic_printf("driver:%s version:%s fw_version:%s bus_info:%s\n",
+                  et_drvinfo->driver, et_drvinfo->version, et_drvinfo->fw_version,
+                  et_drvinfo->bus_info);
+    #endif
+  }
+
+  ALLOCATE_GUARD(et_wolinfo, 'd');
+  GENERIC_ETHTOOL_REQUEST_BY_NAME(et_wolinfo, ETHTOOL_GWOL);
+
+  ALLOCATE_GUARD(et_regs, 'e');
+  et_regs->et.len = sizeof(et_regs->data);
+  GENERIC_ETHTOOL_REQUEST_BY_NAME(&et_regs->et, ETHTOOL_GREGS);
+  if (-1 != ret) {
+    uint32_t i;
+    for (i = 0; i < et_regs->et.len; ++i) {
+      atomic_printf("%02x ", et_regs->data[i]);
+    }
+    atomic_printf("\n");
+  }
+
+  ALLOCATE_GUARD(et_eeprom, 'f');
+  et_eeprom->et.offset = 0;
+  et_eeprom->et.len = sizeof(et_eeprom->data);
+  GENERIC_ETHTOOL_REQUEST_BY_NAME(&et_eeprom->et, ETHTOOL_GEEPROM);
+  if (-1 != ret) {
+    uint32_t i;
+    for (i = 0; i < et_eeprom->et.len; ++i) {
+      atomic_printf("%02x ", et_eeprom->data[i]);
+    }
+    atomic_printf("\n");
+  }
+
+  ALLOCATE_GUARD(et_eeprom, 'g');
+  et_eeprom->et.offset = 0;
+  et_eeprom->et.len = sizeof(et_eeprom->data);
+  GENERIC_ETHTOOL_REQUEST_BY_NAME(&et_eeprom->et, ETHTOOL_GMODULEEEPROM);
+  if (-1 != ret) {
+    uint32_t i;
+    for (i = 0; i < et_eeprom->et.len; ++i) {
+      atomic_printf("%02x ", et_eeprom->data[i]);
+    }
+    atomic_printf("\n");
+  }
+
+  ALLOCATE_GUARD(et_eee, 'h');
+  GENERIC_ETHTOOL_REQUEST_BY_NAME(et_eee, ETHTOOL_GEEE);
+  if (-1 != ret) {
+    atomic_printf("tx_lpi_timer:%d enabled:%d\n",
+                  et_eee->tx_lpi_timer, et_eee->eee_enabled);
+  }
+
+  ALLOCATE_GUARD(et_modinfo, 'i');
+  GENERIC_ETHTOOL_REQUEST_BY_NAME(et_modinfo, ETHTOOL_GMODULEINFO);
+
+  ALLOCATE_GUARD(et_coalesce, 'j');
+  GENERIC_ETHTOOL_REQUEST_BY_NAME(et_coalesce, ETHTOOL_GCOALESCE);
+  if (-1 != ret) {
+    atomic_printf("rx_coalesce_usecs:%d tx_coalesce_usecs:%d\n",
+                  et_coalesce->rx_coalesce_usecs, et_coalesce->tx_coalesce_usecs);
+  }
+
+  ALLOCATE_GUARD(et_ringparam, 'k');
+  GENERIC_ETHTOOL_REQUEST_BY_NAME(et_ringparam, ETHTOOL_GRINGPARAM);
+  if (-1 != ret) {
+    atomic_printf("rx_max_pending:%d rx_pending:%d tx_max_pending:%d tx_pending:%d\n",
+                  et_ringparam->rx_max_pending, et_ringparam->rx_pending,
+                  et_ringparam->rx_pending, et_ringparam->tx_pending);
+  }
+
+  ALLOCATE_GUARD(et_channels, 'l');
+  GENERIC_ETHTOOL_REQUEST_BY_NAME(et_channels, ETHTOOL_GCHANNELS);
+  if (-1 != ret) {
+    atomic_printf("max_rx:%d max_tx:%d\n",
+                  et_channels->max_rx, et_channels->max_tx);
+  }
+
+  ALLOCATE_GUARD(et_pauseparam, 'm');
+  GENERIC_ETHTOOL_REQUEST_BY_NAME(et_pauseparam, ETHTOOL_GPAUSEPARAM);
+  if (-1 != ret) {
+    atomic_printf("rx_pause:%d tx_pause:%d\n",
+                  et_pauseparam->rx_pause, et_pauseparam->tx_pause);
+  }
+
+  ALLOCATE_GUARD(et_sset_info, 'n');
+  et_sset_info->et.sset_mask = 0xff;
+  GENERIC_ETHTOOL_REQUEST_BY_NAME(&et_sset_info->et, ETHTOOL_GSSET_INFO);
+  if (-1 != ret) {
+    int index = 0;
+    for (i = 0; i < 8; ++i) {
+      if (et_sset_info->et.sset_mask & (1 << i)) {
+        uint32_t len = et_sset_info->data[index++];
+        size_t size = sizeof(struct ethtool_gstrings) + len*ETH_GSTRING_LEN;
+        char* buf = (char*)allocate_guard(size, 'o');
+        struct ethtool_gstrings* et_gstrings = (struct ethtool_gstrings*)buf;
+        et_gstrings->cmd = ETHTOOL_GSTRINGS;
+        et_gstrings->string_set = i;
+        req->ifr_data = buf;
+        ret = ioctl(sockfd, SIOCETHTOOL, req);
+        VERIFY_GUARD(req);
+        verify_guard(size, buf);
+        err = errno;
+        atomic_printf("SIOCETHTOOL(ret:%d) ETHTOOL_GSTRINGS: %s ethtool data: ", ret, req->ifr_name);
+        if (-1 == ret) {
+          atomic_printf("WARNING: %s doesn't appear to support SIOCETHTOOL ETHTOOL_GSTRINGS\n", req->ifr_name);
+          test_assert(EOPNOTSUPP == err || EPERM == err);
+        } else {
+          uint32_t j;
+          for (j = 0; j < len; ++j) {
+            atomic_printf("Group %d string %d: %s\n",
+                          i, j, buf + sizeof(struct ethtool_gstrings) + j*ETH_GSTRING_LEN);
+          }
+        }
+      }
+    }
+  }
+
+  ALLOCATE_GUARD(et_gfeatures, 'p');
+  et_gfeatures->et.size = 20;
+  GENERIC_ETHTOOL_REQUEST_BY_NAME(&et_gfeatures->et, ETHTOOL_GFEATURES);
+  if (-1 != ret) {
+    int n = et_gfeatures->et.size;
+    if (n > 20) {
+      n = 20;
+    }
+    for (i = 0; i < n; ++i) {
+      atomic_printf("Feature %d available:%x requested:%x\n",
+        i, et_gfeatures->features[i].available, et_gfeatures->features[i].requested);
+    }
+  }
+
+  ALLOCATE_GUARD(et_perm_addr, 'q');
+  et_perm_addr->et.size = sizeof(et_perm_addr->data);
+  GENERIC_ETHTOOL_REQUEST_BY_NAME(&et_perm_addr->et, ETHTOOL_GPERMADDR);
+  if (-1 != ret) {
+    uint32_t i;
+    for (i = 0; i < et_perm_addr->et.size; ++i) {
+      atomic_printf("%02x ", et_perm_addr->data[i]);
+    }
+    atomic_printf("\n");
+  }
+}
+
 int main(void) {
   int sockfd = socket(AF_INET, SOCK_DGRAM, 0);
   struct ifreq* req;
   struct ifreq* eth_req;
   char name[PATH_MAX];
   int index;
-  struct ethtool_cmd* etc;
   int err, ret;
   struct iwreq* wreq;
   char buf[1024];
@@ -250,23 +457,7 @@ int main(void) {
   GENERIC_REQUEST_BY_NAME(SIOCGIFTXQLEN);
   atomic_printf("qlen is %d\n", req->ifr_qlen);
 
-  ALLOCATE_GUARD(etc, 'b');
-  etc->cmd = ETHTOOL_GSET;
-  req->ifr_data = (char*)&etc;
-  ret = ioctl(sockfd, SIOCETHTOOL, req);
-  VERIFY_GUARD(req);
-  VERIFY_GUARD(etc);
-  err = errno;
-  atomic_printf("SIOCETHTOOL(ret:%d): %s ethtool data: ", ret, req->ifr_name);
-  if (-1 == ret) {
-    atomic_printf("WARNING: %s doesn't appear to support SIOCETHTOOL\n", name);
-    test_assert(EOPNOTSUPP == err || EPERM == err);
-  } else {
-    atomic_printf("speed:%#x duplex:%#x port:%#x physaddr:%#x, maxtxpkt:%u "
-                  "maxrxpkt:%u ...\n",
-                  ethtool_cmd_speed(etc), etc->duplex, etc->port,
-                  etc->phy_address, etc->maxtxpkt, etc->maxrxpkt);
-  }
+  ethtool(sockfd, eth_req);
 
   GENERIC_WIRELESS_REQUEST_BY_NAME(SIOCGIWNAME,
                                    ("wireless protocol name:%s", wreq->u.name));
@@ -292,8 +483,9 @@ int main(void) {
   atomic_printf("SIOCGIWESSID(ret:%d): %s: ", ret, wreq->ifr_name);
   if (-1 == ret) {
     atomic_printf("WARNING: %s doesn't appear to be a wireless iface\n", name);
-    /* "bond" network devices can return ENODEV. */
-    test_assert(EOPNOTSUPP == err || EPERM == err || EINVAL == err || ENODEV == err);
+    /* "bond" network devices can return ENODEV.
+       Some virtual ethernet devices return ENOTTY. */
+    test_assert(EOPNOTSUPP == err || EPERM == err || EINVAL == err || ENODEV == err || ENOTTY == err);
   } else {
     atomic_printf("wireless ESSID:%s\n", buf);
   }
