@@ -5,7 +5,7 @@ expect_gdb('Breakpoint 1')
 send_gdb('c')
 expect_gdb('Breakpoint 1')
 
-send_gdb('p strlen(0)')
+send_gdb('p (int)strdup(0)')
 expect_gdb('received signal SIGSEGV')
 send_gdb('c')
 expect_rr('EXIT-SUCCESS')
