@@ -10,6 +10,10 @@ __asm__ ("my_syscall: mov %rdi,%rax\n\t"
 __asm__ ("my_syscall: mov 4(%esp),%eax\n\t"
          "syscall_instruction: int $0x80\n\t"
          "ret");
+#elif defined(__aarch64__)
+__asm__ ("my_syscall: mov x8, x0\n\t"
+         "svc #0\n\t"
+         "ret");
 #else
 #error define syscall here
 #endif
