@@ -3,6 +3,7 @@
 #include "util.h"
 
 #include <arpa/inet.h>
+#include <dirent.h>
 #include <elf.h>
 #include <execinfo.h>
 #include <fcntl.h>
@@ -11,6 +12,7 @@
 #include <linux/capability.h>
 #include <linux/magic.h>
 #include <linux/prctl.h>
+#include <math.h>
 #include <netinet/in.h>
 #include <stdlib.h>
 #include <string.h>
@@ -24,6 +26,8 @@
 #include <algorithm>
 #include <fstream>
 #include <string>
+#include <numeric>
+#include <random>
 
 #include "preload/preload_interface.h"
 
@@ -32,6 +36,7 @@
 #include "Flags.h"
 #include "PerfCounters.h"
 #include "ReplaySession.h"
+#include "RecordTask.h"
 #include "ReplayTask.h"
 #include "TraceStream.h"
 #include "core.h"
@@ -59,7 +64,7 @@ template <typename Arch> static remote_ptr<typename Arch::unsigned_word> env_ptr
   return stack_ptr;
 }
 
-template <typename Arch> static vector<uint8_t> read_auxv_arch(Task* t) {
+template <typename Arch> remote_ptr<typename Arch::unsigned_word> auxv_ptr(Task* t) {
   auto stack_ptr = env_ptr<Arch>(t);
 
   // Should now point to envp
@@ -68,7 +73,11 @@ template <typename Arch> static vector<uint8_t> read_auxv_arch(Task* t) {
   }
   stack_ptr++;
   // should now point to ELF Auxiliary Table
+  return stack_ptr;
+}
 
+template <typename Arch>
+static vector<uint8_t> read_auxv_arch(Task* t, remote_ptr<typename Arch::unsigned_word> stack_ptr) {
   vector<uint8_t> result;
   while (true) {
     auto pair_vec = t->read_mem(stack_ptr, 2);
@@ -83,10 +92,35 @@ template <typename Arch> static vector<uint8_t> read_auxv_arch(Task* t) {
   return result;
 }
 
+template <typename Arch> static vector<uint8_t> read_auxv_arch(Task* t) {
+  auto stack_ptr = auxv_ptr<Arch>(t);
+  return read_auxv_arch<Arch>(t, stack_ptr);
+}
+
 vector<uint8_t> read_auxv(Task* t) {
   RR_ARCH_FUNCTION(read_auxv_arch, t->arch(), t);
 }
 
+template <typename Arch> void patch_auxv_vdso_arch(RecordTask* t) {
+  auto stack_ptr = auxv_ptr<Arch>(t);
+  std::vector<uint8_t> v = read_auxv_arch<Arch>(t, stack_ptr);
+  size_t wsize = sizeof(typename Arch::unsigned_word);
+  for (int i = 0; (i + 1)*wsize*2 <= v.size(); ++i) {
+    if (*((typename Arch::unsigned_word*)(v.data() + i*2*wsize)) == AT_SYSINFO_EHDR) {
+      auto entry_ptr = stack_ptr + i*2;
+      typename Arch::unsigned_word new_entry = AT_IGNORE;
+      t->write_mem(entry_ptr, new_entry);
+      t->record_local(entry_ptr, &new_entry);
+      return;
+    }
+  }
+  return;
+}
+
+void patch_auxv_vdso(RecordTask* t) {
+  RR_ARCH_FUNCTION(patch_auxv_vdso_arch, t->arch(), t);
+}
+
 template <typename Arch> static vector<string> read_env_arch(Task* t) {
   auto stack_ptr = env_ptr<Arch>(t);
 
@@ -307,7 +341,7 @@ static bool checksum_segment_filter(const AddressSpace::Mapping& m) {
   static const char mmap_clone[] = "mmap_clone_";
   may_diverge =
       m.map.fsname().substr(0, array_length(mmap_clone) - 1) != mmap_clone &&
-      (should_copy_mmap_region(m.map, st) || (PROT_WRITE & m.map.prot()));
+      (should_copy_mmap_region(m.map, m.map.fsname(), st) || (PROT_WRITE & m.map.prot()));
   LOG(debug) << (may_diverge ? "CHECKSUMMING" : "  skipping") << " '"
              << m.map.fsname() << "'";
   return may_diverge;
@@ -503,6 +537,11 @@ static void iterate_checksums(Task* t, ChecksumMode mode,
 }
 
 bool should_checksum(const Event& event, FrameTime time) {
+  FrameTime checksum = Flags::get().checksum;
+  if (Flags::CHECKSUM_NONE == checksum) {
+    return false;
+  }
+
   if (event.type() == EV_EXIT) {
     // Task is dead, or at least detached, and we can't read its memory safely.
     return false;
@@ -513,18 +552,13 @@ bool should_checksum(const Event& event, FrameTime time) {
     return false;
   }
 
-  FrameTime checksum = Flags::get().checksum;
-  bool is_syscall_exit =
-      EV_SYSCALL == event.type() && EXITING_SYSCALL == event.Syscall().state;
-
-  if (Flags::CHECKSUM_NONE == checksum) {
-    return false;
-  }
   if (Flags::CHECKSUM_ALL == checksum) {
-    return true;
+    // Don't checksum at syscall entry because we sometimes mutate syscall parameters there
+    // and that will cause false divergence
+    return EV_SYSCALL != event.type() || ENTERING_SYSCALL != event.Syscall().state;
   }
   if (Flags::CHECKSUM_SYSCALL == checksum) {
-    return is_syscall_exit;
+    return EV_SYSCALL == event.type() && EXITING_SYSCALL == event.Syscall().state;
   }
   /* |checksum| is a global time point. */
   return checksum <= time;
@@ -623,24 +657,20 @@ SignalDeterministic is_deterministic_signal(Task* t) {
   }
 }
 
-static bool has_fs_name(const string& path) {
-  struct stat dummy;
-  return 0 == stat(path.c_str(), &dummy);
-}
-
-static bool is_tmp_file(const string& path) {
+static bool is_tmp_file(const ScopedFd &fd, const string &tracee_path) {
   if (getenv("RR_TRUST_TEMP_FILES")) {
     return false;
   }
   struct statfs sfs;
-  statfs(path.c_str(), &sfs);
-  return (TMPFS_MAGIC == sfs.f_type
+  int ret = fstatfs(fd.get(), &sfs);
+  return ((ret == 0 && TMPFS_MAGIC == sfs.f_type)
           // In observed configurations of Ubuntu 13.10, /tmp is
           // a folder in the / fs, not a separate tmpfs.
-          || path.c_str() == strstr(path.c_str(), "/tmp/"));
+          || tracee_path.c_str() == strstr(tracee_path.c_str(), "/tmp/"));
 }
 
 bool should_copy_mmap_region(const KernelMapping& mapping,
+                             const string &file_name,
                              const struct stat& stat) {
   if (getenv("RR_COPY_ALL_FILES")) {
     return true;
@@ -648,14 +678,21 @@ bool should_copy_mmap_region(const KernelMapping& mapping,
 
   int flags = mapping.flags();
   int prot = mapping.prot();
-  const string& file_name = mapping.fsname();
   bool private_mapping = (flags & MAP_PRIVATE);
 
+  // file_name may point into proc, since we may not have direct access to
+  // the file. Open it, so we can perform the various queries we'd like to ask.
+  ScopedFd fd(file_name.c_str(), O_RDONLY);
+
   // TODO: handle mmap'd files that are unlinked during
   // recording or otherwise not available.
-  if (!has_fs_name(file_name)) {
+  if (!fd.is_open()) {
     // This includes files inaccessible because the tracee is using a different
     // mount namespace with its own mounts
+    // It's also possible for a tracee to mmap a file it doesn't have permission
+    // to read, e.g. if a daemon opened the file and passed the fd over a
+    // socket. We should copy the data now because we won't be able to read
+    // it later. nscd does this.
     LOG(debug) << "  copying unlinked/inaccessible file";
     return true;
   }
@@ -663,18 +700,18 @@ bool should_copy_mmap_region(const KernelMapping& mapping,
     LOG(debug) << "  copying non-regular-file";
     return true;
   }
-  if (is_tmp_file(file_name)) {
+  if (is_tmp_file(fd, mapping.fsname())) {
     LOG(debug) << "  copying file on tmpfs";
     return true;
   }
-  if (file_name == "/etc/ld.so.cache") {
+  if (mapping.fsname() == "/etc/ld.so.cache") {
     // This file changes on almost every system update so we should copy it.
-    LOG(debug) << "  copying " << file_name;
+    LOG(debug) << "  copying " << mapping.fsname();
     return true;
   }
   if (private_mapping && (prot & PROT_EXEC)) {
     /* Be optimistic about private executable mappings */
-    LOG(debug) << "  (no copy for +x private mapping " << file_name << ")";
+    LOG(debug) << "  (no copy for +x private mapping " << mapping.fsname() << ")";
     return false;
   }
   if (private_mapping && (0111 & stat.st_mode)) {
@@ -683,17 +720,9 @@ bool should_copy_mmap_region(const KernelMapping& mapping,
      * Since we're already assuming those change very
      * infrequently, we can avoid copying the data
      * sections too. */
-    LOG(debug) << "  (no copy for private mapping of +x " << file_name << ")";
+    LOG(debug) << "  (no copy for private mapping of +x " << mapping.fsname() << ")";
     return false;
   }
-  bool can_read_file = (0 == access(file_name.c_str(), R_OK));
-  if (!can_read_file) {
-    // It's possible for a tracee to mmap a file it doesn't have permission
-    // to read, e.g. if a daemon opened the file and passed the fd over a
-    // socket. We should copy the data now because we won't be able to read
-    // it later. nscd does this.
-    return true;
-  }
 
   // TODO: using "can the euid of the rr process write this
   // file" as an approximation of whether the tracee can write
@@ -762,7 +791,11 @@ bool xsave_enabled() {
   return (features.ecx & OSXSAVE_FEATURE_FLAG) != 0;
 }
 
+
+#define FATAL_X86_ONLY() FATAL() << "Reached x86-only code on non-x86 arch"
+
 uint64_t xcr0() {
+#if defined(__i386__) || defined(__x86_64__)
   if (!xsave_enabled()) {
     // Assume x87/SSE enabled.
     return 3;
@@ -772,8 +805,13 @@ uint64_t xcr0() {
                : "=a"(eax), "=d"(edx)
                : "c"(0));
   return (uint64_t(edx) << 32) | eax;
+#else
+  FATAL_X86_ONLY();
+  return 0;
+#endif
 }
 
+#if defined(__i386__) || defined(__x86_64__)
 CPUIDData cpuid(uint32_t code, uint32_t subrequest) {
   CPUIDData result;
   asm volatile("cpuid"
@@ -782,21 +820,30 @@ CPUIDData cpuid(uint32_t code, uint32_t subrequest) {
                : "a"(code), "c"(subrequest));
   return result;
 }
+#else
+CPUIDData cpuid(uint32_t, uint32_t) {
+  FATAL_X86_ONLY();
+  __builtin_unreachable();
+}
+#endif
+
 
 #define SEGV_HANDLER_MAGIC 0x98765432
 
 static void cpuid_segv_handler(__attribute__((unused)) int sig,
                                __attribute__((unused)) siginfo_t* si,
                                void* user) {
-  ucontext_t* ctx = (ucontext_t*)user;
 #if defined(__i386__)
+  ucontext_t* ctx = (ucontext_t*)user;
   ctx->uc_mcontext.gregs[REG_EIP] += 2;
   ctx->uc_mcontext.gregs[REG_EAX] = SEGV_HANDLER_MAGIC;
 #elif defined(__x86_64__)
+  ucontext_t* ctx = (ucontext_t*)user;
   ctx->uc_mcontext.gregs[REG_RIP] += 2;
   ctx->uc_mcontext.gregs[REG_RAX] = SEGV_HANDLER_MAGIC;
 #else
-#error unknown architecture
+  (void)user;
+  FATAL_X86_ONLY();
 #endif
 }
 
@@ -1016,6 +1063,44 @@ bool cpuid_compatible(const vector<CPUIDRecord>& trace_records) {
   return cpu_type == trace_cpu_type;
 }
 
+bool cpu_has_xsave_fip_fdp_quirk() {
+#if defined(__i386__) || defined(__x86_64__)
+  uint64_t xsave_buf[576/sizeof(uint64_t)] __attribute__((aligned(64)));
+  xsave_buf[1] = 0;
+  asm volatile("finit\n"
+               "fld1\n"
+#if defined(__x86_64__)
+               "xsave64 %0\n"
+#else
+               "xsave %0\n"
+#endif
+               : "=m"(xsave_buf)
+               : "a"(1), "d"(0)
+               : "memory");
+  return !xsave_buf[1];
+#else
+  FATAL_X86_ONLY();
+  return false;
+#endif
+}
+
+bool cpu_has_fdp_exception_only_quirk() {
+#if defined(__i386__) || defined(__x86_64__)
+  uint32_t fenv_buf[7];
+  fenv_buf[5] = 0;
+  asm volatile("finit\n"
+               "fld1\n"
+               "fstenv %0\n"
+               : "=m"(fenv_buf)
+               :
+               : "memory");
+  return !fenv_buf[5];
+#else
+  FATAL_X86_ONLY();
+  return false;
+#endif
+}
+
 template <typename Arch>
 static CloneParameters extract_clone_parameters_arch(const Registers& regs) {
   CloneParameters result;
@@ -1034,7 +1119,7 @@ static CloneParameters extract_clone_parameters_arch(const Registers& regs) {
   int flags = (int)regs.arg1();
   // If these flags aren't set, the corresponding clone parameters may be
   // invalid pointers, so make sure they're ignored.
-  if (!(flags & CLONE_PARENT_SETTID)) {
+  if (!(flags & (CLONE_PARENT_SETTID | CLONE_PIDFD))) {
     result.ptid = nullptr;
   }
   if (!(flags & (CLONE_CHILD_SETTID | CLONE_CHILD_CLEARTID))) {
@@ -1131,7 +1216,17 @@ double monotonic_now_sec() {
   return (double)tp.tv_sec + (double)tp.tv_nsec / 1e9;
 }
 
-bool running_under_rr() { return getenv("RUNNING_UNDER_RR") != NULL; }
+bool running_under_rr(bool cache) {
+  static bool rr_under_rr = false;
+  static bool rr_check_done = false;
+  if (!rr_check_done || !cache) {
+    rr_check_done = true;
+    int ret = syscall(SYS_rrcall_check_presence, 0, 0, 0, 0, 0, 0);
+    DEBUG_ASSERT(ret == 0 || (ret == -1 && errno == ENOSYS));
+    rr_under_rr = ret == 0;
+  }
+  return rr_under_rr;
+}
 
 vector<string> read_proc_status_fields(pid_t tid, const char* name,
                                        const char* name2, const char* name3) {
@@ -1173,11 +1268,6 @@ vector<string> read_proc_status_fields(pid_t tid, const char* name,
   return result;
 }
 
-bool is_zombie_process(pid_t pid) {
-  auto state = read_proc_status_fields(pid, "State");
-  return state.empty() || state[0].empty() || state[0][0] == 'Z';
-}
-
 static bool check_for_pax_kernel() {
   auto results = read_proc_status_fields(getpid(), "PaX");
   return !results.empty();
@@ -1188,8 +1278,41 @@ bool uses_invisible_guard_page() {
   return !is_pax_kernel;
 }
 
+static bool try_copy_file_by_copy_all(int dest_fd, int src_fd)
+{
+  static bool should_try_copy_all = true;
+  if (!should_try_copy_all) {
+    return false;
+  }
+  struct stat src_stat;
+  if (0 != fstat(src_fd, &src_stat)) {
+    return false;
+  }
+  size_t remaining_size = src_stat.st_size;
+  while (remaining_size > 0) {
+    ssize_t ncopied = syscall(NativeArch::copy_file_range, src_fd, NULL,
+                              dest_fd, NULL, remaining_size, 0);
+    if (ncopied == -1) {
+      if (errno == ENOSYS) {
+        should_try_copy_all = false;
+      }
+      return false;
+    }
+    if (ncopied == 0) {
+      // This doesn't necessarily mean EOF - some file systems don't like this
+      // system call.
+      return false;
+    }
+    remaining_size -= ncopied;
+  }
+  return true;
+}
+
 bool copy_file(int dest_fd, int src_fd) {
   char buf[32 * 1024];
+  if (try_copy_file_by_copy_all(dest_fd, src_fd)) {
+    return true;
+  }
   while (1) {
     ssize_t bytes_read = read(src_fd, buf, sizeof(buf));
     if (bytes_read < 0) {
@@ -1336,7 +1459,7 @@ void dump_rr_stack() {
 }
 
 void check_for_leaks() {
-  if (getenv("RUNNING_UNDER_RR")) {
+  if (running_under_rr()) {
     // Don't do leak checking. The outer rr may have injected maps into our
     // address space that look like leaks to us.
     return;
@@ -1412,6 +1535,42 @@ TempFile create_temporary_file(const char* pattern) {
   return result;
 }
 
+static ScopedFd create_memfd_file(const string &real_name) {
+  ScopedFd fd(syscall(SYS_memfd_create, real_name.c_str(), 0));
+  return fd;
+}
+
+static void replace_char(string& s, char c, char replacement) {
+  size_t i = 0;
+  while (string::npos != (i = s.find(c, i))) {
+    s[i] = replacement;
+  }
+}
+
+// Used only when memfd_create is not available, i.e. Linux < 3.17
+static ScopedFd create_tmpfs_file(const string &real_name) {
+  std::string name = real_name;
+  replace_char(name, '/', '\\');
+  name = string(tmp_dir()) + '/' + name;
+  name = name.substr(0, 255);
+
+  ScopedFd fd =
+      open(name.c_str(), O_CREAT | O_EXCL | O_RDWR | O_CLOEXEC, 0700);
+  /* Remove the fs name so that we don't have to worry about
+   * cleaning up this segment in error conditions. */
+  unlink(name.c_str());
+  return fd;
+}
+
+ScopedFd open_memory_file(const std::string &name)
+{
+  ScopedFd fd(create_memfd_file(name));
+  if (!fd.is_open()) {
+    fd = create_tmpfs_file(name);
+  }
+  return fd;
+}
+
 void good_random(void* out, size_t out_len) {
   ScopedFd fd("/dev/urandom", O_RDONLY);
   uint8_t* o = static_cast<uint8_t*>(out);
@@ -1448,27 +1607,42 @@ static const uint8_t rdtsc_insn[] = { 0x0f, 0x31 };
 static const uint8_t rdtscp_insn[] = { 0x0f, 0x01, 0xf9 };
 static const uint8_t cpuid_insn[] = { 0x0f, 0xa2 };
 static const uint8_t int3_insn[] = { 0xcc };
+static const uint8_t pushf_insn[] = { 0x9c };
+static const uint8_t pushf16_insn[] = { 0x66, 0x9c };
 
+// XXX this probably needs to be extended to decode ignored prefixes
 TrappedInstruction trapped_instruction_at(Task* t, remote_code_ptr ip) {
   uint8_t insn[sizeof(rdtscp_insn)];
-  ssize_t len =
+  ssize_t ret =
       t->read_bytes_fallible(ip.to_data_ptr<uint8_t>(), sizeof(insn), insn);
-  if ((size_t)len >= sizeof(rdtsc_insn) &&
+  if (ret < 0) {
+    return TrappedInstruction::NONE;
+  }
+  size_t len = ret;
+  if (len >= sizeof(rdtsc_insn) &&
       !memcmp(insn, rdtsc_insn, sizeof(rdtsc_insn))) {
     return TrappedInstruction::RDTSC;
   }
-  if ((size_t)len >= sizeof(rdtscp_insn) &&
+  if (len >= sizeof(rdtscp_insn) &&
       !memcmp(insn, rdtscp_insn, sizeof(rdtscp_insn))) {
     return TrappedInstruction::RDTSCP;
   }
-  if ((size_t)len >= sizeof(cpuid_insn) &&
+  if (len >= sizeof(cpuid_insn) &&
       !memcmp(insn, cpuid_insn, sizeof(cpuid_insn))) {
     return TrappedInstruction::CPUID;
   }
-  if ((size_t)len >= sizeof(int3_insn) &&
+  if (len >= sizeof(int3_insn) &&
       !memcmp(insn, int3_insn, sizeof(int3_insn))) {
     return TrappedInstruction::INT3;
   }
+  if (len >= sizeof(pushf_insn) &&
+      !memcmp(insn, pushf_insn, sizeof(pushf_insn))) {
+    return TrappedInstruction::PUSHF;
+  }
+  if (len >= sizeof(pushf16_insn) &&
+      !memcmp(insn, pushf16_insn, sizeof(pushf16_insn))) {
+    return TrappedInstruction::PUSHF16;
+  }
   return TrappedInstruction::NONE;
 }
 
@@ -1481,35 +1655,60 @@ size_t trapped_instruction_len(TrappedInstruction insn) {
     return sizeof(cpuid_insn);
   } else if (insn == TrappedInstruction::INT3) {
     return sizeof(int3_insn);
+  } else if (insn == TrappedInstruction::PUSHF) {
+    return sizeof(pushf_insn);
+  } else if (insn == TrappedInstruction::PUSHF16) {
+    return sizeof(pushf16_insn);
   } else {
     return 0;
   }
 }
 
+/**
+ * Certain instructions generate deterministic signals but also advance pc.
+ * Look *backwards* and see if this was one of them.
+ */
+bool is_advanced_pc_and_signaled_instruction(Task* t, remote_code_ptr ip) {
+#if defined(__i386__) || defined(__x86_64__)
+  uint8_t insn[sizeof(int3_insn)];
+  ssize_t ret =
+      t->read_bytes_fallible(ip.to_data_ptr<uint8_t>() - sizeof(insn), sizeof(insn), insn);
+  if (ret < 0) {
+    return false;
+  }
+  size_t len = ret;
+  if (len >= sizeof(int3_insn) &&
+      !memcmp(insn, int3_insn, sizeof(int3_insn))) {
+    return true;
+  }
+#endif
+  return false;
+}
+
 /**
  * Read and parse the available CPU list then select a random CPU from the list.
  */
-static int get_random_cpu_cgroup() {
+static vector<int> get_cgroup_cpus() {
+  vector<int> cpus;
   ifstream self_cpuset("/proc/self/cpuset");
   if (!self_cpuset.is_open()) {
-    return -1;
+    return cpus;
   }
   string cpuset_path;
   getline(self_cpuset, cpuset_path);
   self_cpuset.close();
   if (cpuset_path.empty()) {
-    return -1;
+    return cpus;
   }
   ifstream cpuset("/sys/fs/cgroup/cpuset" + cpuset_path + "/cpuset.cpus");
   if (!cpuset.good()) {
-    return -1;
+    return cpus;
   }
-  vector<int> cpus;
   while (true) {
     int cpu1;
     cpuset >> cpu1;
     if (cpuset.fail()) {
-      return -1;
+      return std::vector<int>{};
     }
     cpus.push_back(cpu1);
     char c = cpuset.get();
@@ -1518,12 +1717,12 @@ static int get_random_cpu_cgroup() {
     } else if (c == ',') {
       continue;
     } else if (c != '-') {
-      return -1;
+      return std::vector<int>{};;
     }
     int cpu2;
     cpuset >> cpu2;
     if (cpuset.fail()) {
-      return -1;
+      return std::vector<int>{};
     }
     for (int cpu = cpu1 + 1; cpu <= cpu2; cpu++) {
       cpus.push_back(cpu);
@@ -1532,21 +1731,53 @@ static int get_random_cpu_cgroup() {
     if (cpuset.eof() || c == '\n') {
       break;
     } else if (c != ',') {
-      return -1;
+      return std::vector<int>{};
     }
   }
-  return cpus[random() % cpus.size()];
+  return cpus;
+}
+
+static string get_cpu_lock_file() {
+  const char* lock_file = getenv("_RR_CPU_LOCK_FILE");
+  return lock_file ? lock_file : trace_save_dir() + "/cpu_lock";
 }
 
 /**
  * Pick a CPU at random to bind to, unless --cpu-unbound has been given,
  * in which case we return -1.
  */
-int choose_cpu(BindCPU bind_cpu) {
+int choose_cpu(BindCPU bind_cpu, ScopedFd &cpu_lock_fd_out) {
   if (bind_cpu == UNBOUND_CPU) {
     return -1;
   }
 
+  // Find out which CPUs we're allowed to run on at all
+  std::vector<int> cpus = get_cgroup_cpus();
+  if (cpus.empty()) {
+    cpus.resize(get_num_cpus());
+    std::iota(cpus.begin(), cpus.end(), 0);
+  }
+
+  // When many copies of rr are running on the same machine, it's easy for them
+  // to oversubscribe a CPU. To avoid this situation, we have a lock file, where
+  // each running rr process will lock the n-th byte if it is running on that
+  // particular CPU. That way subsequent rr processes can avoid in-use cores
+  // if possible.
+  string cpu_lock_file = get_cpu_lock_file();
+  cpu_lock_fd_out = ScopedFd(cpu_lock_file.c_str(), O_CREAT | O_WRONLY | O_CLOEXEC, 0600);
+
+  // Make sure the locks file is big enough
+  if (cpu_lock_fd_out.is_open()) {
+    struct stat stat;
+    int err = fstat(cpu_lock_fd_out, &stat);
+    DEBUG_ASSERT(err == 0);
+    if (stat.st_size < get_num_cpus()) {
+      if (ftruncate(cpu_lock_fd_out, get_num_cpus())) {
+        FATAL() << "Failed to resize locks file";
+      }
+    }
+  }
+
   // Pin tracee tasks to a random logical CPU, both in
   // recording and replay.  Tracees can see which HW
   // thread they're running on by asking CPUID, and we
@@ -1562,14 +1793,50 @@ int choose_cpu(BindCPU bind_cpu) {
   // presumably due to cheaper context switching and/or
   // better interaction with CPU frequency scaling.
   if (bind_cpu >= 0) {
+    if (cpu_lock_fd_out.is_open()) {
+      struct flock lock {
+        .l_type = F_WRLCK,
+        .l_whence = SEEK_SET,
+        .l_start = bind_cpu,
+        .l_len = 1,
+        .l_pid = 0
+      };
+      // Try to acquire the lock for this CPU
+      (void)fcntl(cpu_lock_fd_out, F_SETLK, &lock);
+      // Ignore fcntl errors - nothing we can do
+    }
     return bind_cpu;
   }
 
-  int cpu = get_random_cpu_cgroup();
-  if (cpu >= 0) {
-    return cpu;
+  if (cpu_lock_fd_out.is_open()) {
+    // Try twice to allocate a CPU. If we fail twice, pick a random one
+    for (int i = 0; i < 2; ++i) {
+      std::shuffle (cpus.begin(), cpus.end(), std::default_random_engine(random()));
+      for (int cpu : cpus) {
+        struct flock lock {
+          .l_type = F_WRLCK,
+          .l_whence = SEEK_SET,
+          .l_start = cpu,
+          .l_len = 1,
+          .l_pid = 0
+        };
+        // Try to acquire the lock for this CPU
+        int err = fcntl(cpu_lock_fd_out, F_SETLK, &lock);
+        if (err == 0) {
+          return cpu;
+        }
+        else if (err == -1) {
+          if (errno != EACCES && errno != EAGAIN) {
+            FATAL() << "Unexpected error trying to acquire CPU lock";
+          }
+        }
+      }
+    }
   }
-  return random() % get_num_cpus();
+
+  // Didn't work - just use a random CPU
+  cpu_lock_fd_out.close();
+  return cpus[random() % cpus.size()];
 }
 
 uint32_t crc32(uint32_t crc, unsigned char* buf, size_t len) {
@@ -1636,7 +1903,7 @@ void write_all(int fd, const void* buf, size_t size) {
   }
 }
 
-ssize_t pwrite_all_fallible(int fd, const void* buf, size_t size, off_t offset) {
+ssize_t pwrite_all_fallible(int fd, const void* buf, size_t size, off64_t offset) {
   ssize_t written = 0;
   while (size > 0) {
     ssize_t ret = ::pwrite64(fd, buf, size, offset);
@@ -1732,4 +1999,87 @@ string json_escape(const string& str, size_t pos) {
   return out;
 }
 
+void sleep_time(double t) {
+  struct timespec ts;
+  ts.tv_sec = (time_t)floor(t);
+  ts.tv_nsec = (long)((t - ts.tv_sec) * 1e9);
+  nanosleep(&ts, NULL);
+}
+
+static bool is_component(const char* p, const char* component) {
+  while (*component) {
+    if (*p != *component) {
+      return false;
+    }
+    ++p;
+    ++component;
+  }
+  return *p == '/' || !*p;
+}
+
+void normalize_file_name(string& s)
+{
+  size_t s_len = s.size();
+  size_t out = 0;
+  for (size_t i = 0; i < s_len; ++i) {
+    if (s[i] == '/') {
+      if (s.c_str()[i + 1] == '/') {
+        // Skip redundant '/'
+        continue;
+      }
+      if (is_component(s.c_str() + i + 1, ".")) {
+        // Skip redundant '/.'
+        ++i;
+        continue;
+      }
+      if (is_component(s.c_str() + i + 1, "..")) {
+        // Peel off '/..'
+        size_t p = s.rfind('/', out - 1);
+        if (p != string::npos) {
+          out = p;
+          i += 2;
+          continue;
+        }
+      }
+    }
+    s[out] = s[i];
+    ++out;
+  }
+  s.resize(out);
+}
+
+std::vector<int> read_all_proc_fds(pid_t tid)
+{
+  std::vector<int> ret;
+  char buf[1000];
+  sprintf(buf, "/proc/%d/fd", tid);
+  DIR *fddir = opendir(buf);
+  DEBUG_ASSERT(fddir != nullptr);
+  while (struct dirent *dir = readdir(fddir)) {
+    ret.push_back(atoi(dir->d_name));
+  }
+  return ret;
+}
+
+std::string find_exec_stub(SupportedArch arch) {
+  string exe_path = resource_path() + "bin/";
+  if (arch == x86 && NativeArch::arch() == x86_64) {
+    exe_path += "rr_exec_stub_32";
+  } else {
+    exe_path += "rr_exec_stub";
+  }
+  return exe_path;
+}
+
+int pop_count(uint64_t v) {
+  int ret = 0;
+  while (v > 0) {
+    if (v & 1) {
+      ++ret;
+    }
+    v >>= 1;
+  }
+  return ret;
+}
+
 } // namespace rr
